#!/usr/bin/env python
"""
Get files from a remote server (to be reached by nfs, samba, ftp, html or local directory) 

file content is directly added to a data bank (or local file if preferred).
"""

from magpy.stream import *
from magpy.core import database
from magpy.core.methods import testtime
from magpy.opt import cred as mpcred

import getopt
import fnmatch
import pexpect
import ftplib
import pwd
import zipfile
import tempfile
from dateutil import parser
from datetime import datetime, timezone
from shutil import copyfile
import filecmp
import subprocess
import socket

from martas.core.methods import martaslog as ml
from martas.core import methods as mm

"""
DESCRIPTION
    Downloads data by default in to an archive "raw" structure
    like /srv/archive/STATIONID/SENSORID/raw
    Adds data into a MagPy database (if writedatabase is True)
    Adds data into a basic archive structure (if writearchive is True)
    The application requires credentials of remote source and local database created by addcred

APPLICATION

   1) Getting binary data from a FTP Source every, scheduled day
    python3 file_download.py -c ../conf/collect-ftpsource.cfg
    in config "collect-ftpsource.cfg":
             sourcedatapath        :      /remote/data
             filenamestructure     :      *%s.bin

   2) Getting binary data from a FTP Source every, scheduled day, using seconday time column and an offset of 2.3 seconds
    python3 file_download -c ../conf/collect-ftpsource.cfg
    in config "collect-ftpsource.cfg":
             sourcedatapath        :      /remote/data
             filenamestructure     :      *%s.bin
             LEMI025_28_0002       :      defaulttimecolumn:sectime;time:-2.3

   3) Just download raw data to archive
    python3 file_download -c ../conf/collect-ftpsource.cfg
    in config "collect-ftpsource.cfg":
             writedatabase     :      False
             writearchive      :      False

   4) Rsync from a ssh server (passwordless access to remote machine is necessary, cred file does not need to contain a pwd) 
    python3 file_download.py -c ../conf/collect-ftpsource.cfg
    in config "collect-ftpsource.cfg":
             protocol          :      rsync
             writedatabase     :      False
             writearchive      :      False

   5) Uploading raw data from local raw archive
    python3 file_download.py -c ../conf/collect-localsource.cfg
    in config "collect-localsource.cfg":
             protocol          :      
             sourcedatapath    :      /srv/archive/DATA/SENSOR/raw
             writedatabase     :      False
             writearchive      :      True
             forcerevision     :      0001


    python3 file_download.py -c ../conf/collect-source.cfg -d 10 -e 2020-10-20

    Debugging option:
    python3 file_download.py -c ../conf/collect-source.cfg -D




CONFIGURATION FILE
 one for each source e.g. collect-janus.cfg


# Data source
# -----------------
# Credentials
sourcecredentials     :      janus

# Path to the data to be collected
sourcedatapath        :      /srv/mqtt

# Protocol for data access (ftp,rsync,scp)
protocol              :      ftp

# Optional - ID of the sensor (required if not contained in the data')
#sensorid              :      xxx_xxx_0001

# Optional - ID of the station i.e. the Observatory code (required if')
#stationid             :      wic

# Dateformat in files to be read
#   like "%Y-%m-%d" for 2014-02-01
#        "%Y%m%d" for 20140201
#        "ctime" or "mtime" for using timestamp of file
dateformat             :      %Y-%m-%d

# filename of data file to be read.
#    Add %s as placeholder for date
#       examples: "WIC_%s.bin"
#                 "*%s*"
#                 "WIC_%s.*"
                  "*.[bin,asc,cdf]" -> define possible variants within brackets 
                  "CALY*[LFZ,LFY,LFX]*%s*" -> brackets can also be used in filename in combiation with date
#                 "WIC_2013.all" - no dateformat -> single file will be read
filenamestructure      :      *%s*

# Timerange
defaultdepth           :      2

# Sensor specific modifications - defaulttimecolumn, offsets by KEY:value pairs
SENSORID               :      defaulttimecolumn:sectime;sectime:2.3

# Perform as user - uncomment if not used
# necessary for cron and other root jobs 
defaultuser     :      cobs

# Walk through subdirectories
# if selected all subdirectories below remote path will be searched for
# filename pattern. Only works for local directories and scp.
walksubdirs     :      False

# Sensors present in path to be skipped (Begging of Sensorname is enough
#blacklist       :      None


# Collecting server
# -----------------
# Archivepath:
# two subdirectories will be created if not existing - based on stationID and sensorID
# e.g. WIC/LEMI025_22_0003
archivepath            :      /srv/archive
# setting subdirectory will lead to date related subdirectories within raw data folder
# supported are year "Y" i.e. /raw/2024/* or "Ym" i.e. /raw/2024/03/*
#subdirectory    :      Y

# If forcedirectory, then archivepath is used for saving data
forcedirectory     :      False


# Zip data in archive directory
zipdata            :      False


# delete from remote source after successful transfer
# (doesnt work with scp)
deleteremote       :      False

# Force data to the given revision number
#forcerevision      :      0001

# Database (Credentials makes use of addcred)
dbcredentials     :      cobsdb

# Disable proxy settings of the system (seems to be unused - check)
disableproxy       :      False

writedatabase     :      True

# Create a basic archiving file without database if True
# basic path is /STATIONID/SENSORID/SENSORID_0001/
writearchive      :      False
archiveformat     :      PYCDF


# Logging
# -----------------
# Logging parameter
# ################
# path to log file
logpath   :   /var/log/magpy/archivestatus.log
# log,email,telegram
notification    :    telegram
# configuration for notification
notificationconf   :   /myconfpath/mynotificationtype.cfg



Changelog:
2014-08-02:   RL removed break when no data was found (could happen if at this selected day not data is available. All other days need to be collected however.
2014-10-22:   RL updated the description
2014-11-04:   RL added the inserttable option to force data upload to a specific table (e.g. for rcs conrad data which has a variable sampling rate)
2015-10-20:   RL changes for fast ndarrays and zip option
2016-10-10:   RL updated imports, improved help and checked for pure file access
2017-03-10:   RL activated force option
2018-10-22:   RL changed all routines considerably
2020-10-01:   RL included and tested rsync option (not perfect, but well, its working)
2021-02-09:   RL rewriting with config (MARCOS sister script of file_upload) --- file_download with add to MagPy database option, column selectors and header updates (e.g. secondary time + offset option)
   - database writing as option (fixed table or not)
   - simple creation of MagPy archive without database tables (but making use of DB meta info)
   - just download files using ftp, scp, rsync etc
   - use a configuration file
2021-07-17    RL complex placeholders for file identifications

"""


def walk_dir(directory_path, filename, date, dateformat, debug=False):
    """
    Method to extract filename with wildcards or date patterns by walking through a local directory structure
    """
    # Walk through files in directory_path, including subdirectories
    pathlist = []
    tcheck = datetime.now()
    if filename == '':
        filename = '*'
    if dateformat in ['','ctime','mtime']:
        filepat = filename
    else:
        filepat = filename % date
    if debug:
        print ("Checking directory {} for files with {}".format(directory_path, filepat))
    for root, _, filenames in os.walk(directory_path):
        for filename in filenames:
            if fnmatch.fnmatch(filename, filepat):
                file_path = os.path.join(root,filename)
                if dateformat in ['ctime','mtime']:
                    if dateformat == 'ctime':
                        tcheck = datetime.fromtimestamp(os.path.getctime(file_path))
                    if dateformat == 'mtime':
                        tcheck = datetime.fromtimestamp(os.path.getmtime(file_path))
                    if tcheck.date() == date.date():
                        pathlist.append(file_path)
                else:
                    pathlist.append(file_path)
    return pathlist

"""
def list_recursive(ftp, remotedir):
    #""
    Get a recursive file listing on FTP sites
    https://stackoverflow.com/questions/65123271/recursive-file-list-with-ftp
    #""
    ftp.cwd(remotedir)
    for entry in ftp.mlsd():
        if entry[1]['type'] == 'dir':
            remotepath = remotedir + "/" + entry[0]
            print(time.time() - t0, remotepath)
            list_recursive(ftp, remotepath)
        else:
            print(entry)
"""

def dir_extract(lines, filename, date, dateformat):
    """
    Method to extract filename with wildcards or date patterns from a directory listing
    """
    pathlist = []
    if filename == '':
        filename = '*'
    if dateformat in ['','ctime','mtime']:
        filepat = filename
    else:
        filepat = filename % date
    for line in lines:
        #print ("Checking line {}".format(line))
        tokens = line.split()
        # Not interested in directories
        if not tokens[0][0] == "d" and len(tokens)==9:
            time_str = tokens[5] + " " + tokens[6] + " " + tokens[7]
            if dateformat in ['ctime','mtime']:
                # cannot distinguish between mtime and ctime here
                time = parser.parse(time_str)
                #print (time.date())
                if time.date() == date.date():
                    pathlist.append(tokens[8])
            else:
                if fnmatch.fnmatch(tokens[8], filepat):
                    file_path = tokens[8]
                    #print (tokens[8])
                    pathlist.append(file_path)
        elif not tokens[0][0] == "d":
            # exception for Geoelectric (Windows FTP)
            if tokens[2] == '<DIR>':
                # Not interested in directories
                continue
            file_path = tokens[3]
            if len(tokens) == 5:
                # TODO blanks in filenames!!!
                file_path = file_path+' '+tokens[4]
            if fnmatch.fnmatch(file_path, filepat):
                pathlist.append(file_path)
    return pathlist


def die(child, errstr):
    print (errstr)
    print (child.before, child.after)
    child.terminate()
    exit(1)


def ssh_getlist(source, filename, date, dateformat, maxdate, cred=None, pwd_required=True, timeout=60, debug=False):
    """
    Method to extract filename with wildcards or date patterns from a directory listing
    """
    if not cred:
        cred = []
    pathlist = []
    #filename = filename.replace('*','')
    print (filename, date)
    if dateformat in ['','ctime','mtime']:
        filepat = filename
    else:
        filepat = filename % date

    if debug:
        print ("Running ssh_getlist for {}".format(filepat))

    def _obtain_search_criteria(fn):
        # BASIC search string
        # fn = filepat
        #grep = 'grep "%s"'
        #search = 'find %s -type f{}{}'.format(find, grep)
        optstr=""
        grepstr=""
        if fn.find("[") >= 0:
            brack = fn[fn.find("[")+1:fn.find("]")]
            opt = brack.split(',')
            for o in opt:
                if o == opt[-1]:
                    optstr += ' -name "*{}*"'.format(o)
                else:
                    optstr += ' -name "*{}*" -o'.format(o)
            fn = fn.replace(brack,'').replace('[]','*')
        el = fn.split('*')
        el = [e for e in el if len(e) > 0]
        for e in el:
            grepstr += ' | grep "{}"'.format(e)
        return optstr, grepstr

    opt, grep = _obtain_search_criteria(filepat)
    #searchstr = 'find %s -type f{}{}'.format(opt, grep)
    #grep += ' | grep -v "#"' #eventually add this exclude
    now = datetime.now(timezone.utc).replace(tzinfo=None)

    if not dateformat in ['','ctime','mtime']:
        searchstr = 'find {} -type f{}{}'.format(source,opt,grep)
    elif dateformat in ['ctime','mtime']:
        mindate = (now - date).days
        maxdate = (now - maxdate).days
        if maxdate == 0:
            searchstr = 'find {} -type f -{} -{}{}{}'.format(source,dateformat, mindate,opt,grep)
        else:
            searchstr = 'find {} -type f -{} -{} -{} +{}{}{}'.format(source,dateformat,mindate,dateformat,(mindate-maxdate),opt,grep)
    else:
        searchstr = 'find {} -type f{}{}'.format(source,opt,grep)

    if debug:
        print ("Searchstring to indentify files to copy: {}".format(searchstr))

    COMMAND= "ssh %s@%s '%s';" % (cred[0],cred[2],searchstr)
    child = pexpect.spawn(COMMAND)
    result = ""
    if timeout:
        child.timeout=timeout
    if pwd_required:
        i = child.expect([pexpect.TIMEOUT, 'assword: '])
        child.sendline(cred[1])
    i = child.expect([pexpect.TIMEOUT, 'Permission denied', pexpect.EOF])
    if i == 0:
        die(child, 'ERROR!\nSSH timed out. Here is what SSH said:')
    elif i == 1:
        die(child, 'ERROR!\nIncorrect password Here is what SSH said:')
    elif i == 2:
        result = child.before
    if sys.version_info.major == 3:
        result = result.decode('ascii')
    pathlist = result.split('\r\n')
    pathlist = [elem for elem in pathlist if not elem == '' and not elem == ' ']
    return pathlist


def check_configuration(config=None,debug=False):
    """
    DESCRIPTION
        configuration data will be checked
    """
    if not config:
        config = {}

    user = ''
    password = ''
    address = ''
    destination = ''
    source = ''
    port = 21
    db = None
    success = True

    if debug:
        print ("  Checking configuration data")

    creddb =  config.get('dbcredentials','')

    if config.get('archivepath') == '' and creddb == '':
        print('Specify either a shortcut to the credential information of the database or a local path:')
        print('-- check file_download.py -h for more options and requirements')
        success = False
        #sys.exit()
    if config.get('archivepath') == '' or config.get('archivepath') == '/archivepath':
        destination = tempfile.gettempdir()
    else:
        if not os.path.isdir(config.get('archivepath')):
            print ("Destination directory {} not existing. Creating it".format(config.get('archivepath')))
            os.makedirs(config.get('archivepath'))
        destination = config.get('archivepath')
    config['destination'] = destination

    credtransfer = config.get('sourcecredentials')
    if not credtransfer == '':
        if debug:
            print ("   - checking credentials for remote access")
        user=mpcred.lc(credtransfer,'user')
        password=mpcred.lc(credtransfer,'passwd')
        address = mpcred.lc(credtransfer,'address')
        try:
            port = int(mpcred.lc(credtransfer,'port'))
        except:
            port = 21
        if debug:
            print ("   -> done")
    config['rmuser'] = user
    config['rmpassword'] = password
    config['rmaddress'] = address
    config['rmport'] = port

    source = ''
    protocol = config.get('protocol')
    if not protocol in ['','ftp','FTP']:
        source += protocol + "://"
        if not user == '' and not password=='':
            source += user + ":" + password + "@"
        if not address == '':
            source += address

    remotepath = config.get('sourcedatapath')
    if not remotepath == '':
        s = ""
        # if a colon is contained in path (i.e. ftp access on windows)
        if isinstance(remotepath,dict):
            for i in remotepath:
                s = "{}:{}".format(i,remotepath[i])
            remotepath = s
        source += remotepath
    config['source'] = source

    if not protocol in ['','scp','ftp','SCP','FTP','html','rsync']:
        print('Specify a valid protocol:')
        print('-- check file_download.py -h for more options and requirements')
        success = False
        #sys.exit()

    walk = config.get('walksubdirs')
    if debug:
        print ("   Walk through subdirs: {}".format(walk))
    if mm.get_bool(walk):
        if not protocol in ['','scp','rsync']: 
            print('   -> Walk mode only works for local directories and scp access.')
            print('   -> Switching walk mode off.')
            config['walksubdirs'] = False

    if not creddb == '':
        print("   Accessing local data bank ...")
        # required for either writeing to DB or getting meta in case of writing archive
        try:
            db = mm.connect_db(creddb)
        except:
            db = None
    config['db'] = db

    # loaded all credential (if started from root permissions are required for that)
    # now switch user for scp
    # TODO check whether this is working in a function
    if config.get('defaultuser'):
        try:
            uid=pwd.getpwnam(config.get('defaultuser'))[2]
            os.setuid(uid)
        except:
            print ("  User {} not existing -  moving on".format(config.get('defaultuser')))

    dateformat = config.get('dateformat')
    filename = config.get('filenamestructure')
    if isinstance(filename,list):
        # necessary if brackets with options are contained
        filename = ",".join(filename)
    config['filenamestruc'] = filename

    if dateformat == "" and filename == "":
        print('   Specify either a fileformat: -f myformat.dat or a dateformat -d "%Y",ctime !')
        print('   -- check file_download.py -h for more options and requirements')
        success = False
        #sys.exit()
    if not dateformat in ['','ctime','mtime']:
        current = datetime.now(timezone.utc).replace(tzinfo=None)
        try:
            newdate = current.strftime(dateformat)
        except:
            print('   Specify a vaild datetime dateformat like "%Y-%m-%d"')
            print('   -- check file_download.py -h for more options and requirements')
            success = False
            #sys.exit()
    if "%s" in filename and dateformat in ['','ctime','mtime']:
        print('   Specify a datetime dateformat for given placeholder in fileformat!')
        print('   -- check file_download.py -h for more options and requirements')
        success = False
        #sys.exit()
    elif not "%s" in filename and "*" in filename and not dateformat in ['ctime','mtime']:
        print('   Specify either ctime or mtime for dateformat to be used with your give fileformat!')
        print('   -- check file_download.py -h for more options and requirements')
        success = False
        #sys.exit()
    elif not "%s" in filename and not "*" in filename and not dateformat in [""]:
        print('   Give dateformat will be ignored!')
        print('   -- check file_download.py -h for more options and requirements')
        print('   -- continuing ...')

    if debug:
        print("  => Configuration checked - success")

    return config, success


def get_datelist(config=None,current=None,debug=False):
    if not current:
        current = datetime.now(timezone.utc).replace(tzinfo=None)
    if not config:
        config = {}
    if debug:
        print("   -> Obtaining timerange ...")
    datelist = []
    newcurrent = current
    dateformat = config.get('dateformat')
    depth = int(config.get('defaultdepth'))
    if not dateformat in ['','ctime','mtime']:
        for elem in range(depth):
            if dateformat == '%b%d%y': #exception for MAGREC
                newdate = newcurrent.strftime(dateformat)
                datelist.append(newdate.upper())
            else:
                datelist.append(newcurrent.strftime(dateformat))
            newcurrent = current-timedelta(days=elem+1)
    elif dateformat in ['ctime','mtime']:
        for elem in range(depth):
            datelist.append(newcurrent)
            newcurrent = current-timedelta(days=elem+1)
    else:
        datelist = ['dummy']
    if depth == 0:
        #exception for Geoelectric
        newcurrent = current-timedelta(hours=1)
        datelist = [newcurrent.strftime('%Y-%m-%d %H')]

    #if debug:
    print("   -> Dealing with time range:\n {}".format(datelist))

    return datelist


def create_transfer_list(config=None,datelist=None,debug=False):
    """
    DESCRIPTION
        Create a list of files to be transfered
        Define source based on 'protocol', 'remotepath', 'walk'
        protocols: ''(local disk), 'scp', 'ftp', 'html'

        What about rsync? -> no need to use create list but requires passwd-less connection  

    RETURNS
        filelist (a list of remote filepaths to be transferred)
    """

    #filelist = getfilelist(protocol, source, sensorid, filename, datelist, walk=True, option=None)
    if not config:
        config = {}
    if not datelist:
        datelist = []

    if debug:
        print(" Getting filelists")
        print(" -------------------")

    protocol = config.get('protocol','')
    source = config.get('source','')
    remotepath = config.get('sourcedatapath')
    filename = config.get('filenamestruc')
    user = config.get('rmuser')
    password = config.get('rmpassword')
    address = config.get('rmaddress')
    port = config.get('rmport', 21)

    dateformat = config.get('dateformat')

    now = datetime.now(timezone.utc).replace(tzinfo=None)
    filelist = []
    if protocol in ['ftp','FTP']:
        print ("  Connecting to FTP")
        if debug:
            print (" - Getting filelist - by ftp ") 
        import ftplib
        if debug:
            print (" - connecting to {} on port {}".format(address,port)) 
        if not port == 21:
            ftp = ftplib.FTP()
            ftp.connect(address,port)
        else:
            ftp = ftplib.FTP(address)
        if debug:
            print (" - user: {} ".format(user)) 
        ftp.login(user,password)
        ftp.cwd(source)
        lines = []
        ftp.dir("", lines.append)
        ftp.close()
        for date in datelist:
            path = dir_extract(lines, filename, date, dateformat)
            if len(path) > 0:
                filelist.extend(path)
    elif protocol in ['scp','SCP','rsync']:
        if debug:
            print (" Connecting for {}".format(protocol))
        pwd_required=True
        if protocol == 'rsync':
            pwd_required=False
            print ("   Rsync requires passwordless ssh connection to remote system")
        import pexpect
        if not dateformat in ['','ctime','mtime']:
            for date in datelist:
                path = ssh_getlist(remotepath, filename, date, dateformat, now, cred=[user,password,address],pwd_required=pwd_required,debug=debug)
                if len(path) > 0:
                    filelist.extend(path)
        else:
            filelist = ssh_getlist(remotepath, filename, min(datelist), dateformat, max(datelist), cred=[user,password,address],pwd_required=pwd_required,debug=debug)
    elif protocol == '':
        if debug:
            print (" Local directory access ") 
        ### Search local directory - Working
        for date in datelist:
            path = walk_dir(source, filename, date, dateformat,debug=debug)
            if len(path) > 0:
                filelist.extend(path)
    elif protocol == 'html':
        print (filelist)
        print ("  HTML access not supported - use MagPy directly to access webservices")

    #if debug:
    print ("Files to be transferred")
    print ("-----------------------------")
    print (filelist)
    print ("-----------------------------")

    return filelist


def obtain_data_files(config=None,filelist=None,debug=False):
    """
    DESCRIPTION
        Download data files ane write either to raw directory ot tmp (or to specified folder)
    ###   2.3 Get selected files and copy them to destination
    ###
    ### only if not protocol == '' and localpath

    ### update filelist with new filenamens on local harddisk

        What about rsync? -> no need to use create list but requires passwd-less connection

    RETURNS
        localfilelist (a list with full paths to all files copied to the localfilesystem)
    """

    if not config:
        config = {}
    if not filelist:
        filelist = []

    # Requires
    stationid = config.get('stationid')
    # if sensorid is not provided it will be extracted from the filelist
    sensorid = config.get('sensorid')
    localpath = config.get('archivepath')
    protocol = config.get('protocol')
    source = config.get('source')
    destination = config.get('destination')
    usedatedir = config.get('subdirectory','')
    deleteremote = config.get('deleteremote',False)
    user = config.get('rmuser')
    password = config.get('rmpassword')
    address = config.get('rmaddress')
    port = config.get('rmport',21)
    zipping = mm.get_bool(config.get('zipdata'))
    forcelocal = mm.get_bool(config.get('forcedirectory',False))
    deleteopt = " "
    sensid = ""
    ftp = None

    #filename = config.get('filenamestructure')
    #dateformat = config.get('dateformat')

    def createdestinationpath(localpath='', stationid='', sensorid='', datedir='', forcelocal=False):
            subdir = 'raw'
            if not stationid and not sensorid or forcelocal:
                destpath = os.path.join(localpath)
            elif not stationid:
                destpath = os.path.join(localpath,sensorid,'raw')
            elif not sensorid:
                destpath = os.path.join(localpath,stationid.upper())
            else:
                destpath = os.path.join(localpath,stationid.upper(),sensorid,'raw')
            if not datedir == '':
                destpath = os.path.join(destpath, str(datedir))
            return destpath


    print("  Writing data to a local directory (or tmp)")

    if deleteremote in [True,'True']:
        print("  IMPORTANT: deleting remote data has been activated")

    if debug:
        print ("   Please Note: files will be copied to local filesystem even when debug is selected")

    localpathlist = []

    if not protocol == '' or (protocol == '' and not destination == tempfile.gettempdir()):
        ### Create a directory by getting sensorid names (from source directory)
        # Open the specific channel
        if protocol in ['ftp','FTP']:
            if not port == 21:
                ftp = ftplib.FTP()
                ftp.connect(address,port)
            else:
                ftp = ftplib.FTP(address)
            ftp.login(user,password)
            ftp.cwd(source)

        for f in filelist:
            datedir = ''
            if debug:
                print ("   Accessing file {}".format(f))
            if usedatedir in ['Y','Ym','Yj']:
                # get year from filename or ctime/mtime
                tcheck = datetime.fromtimestamp(os.path.getmtime(f))
                fyear = tcheck.year
                fmonth = tcheck.month
                dirdoy = tcheck.strftime('%j')

                if usedatedir == 'Y':
                    # get year from filename or ctime/mtime
                    datedir = fyear
                elif usedatedir == 'Ym':
                    datedir = '{}/{:02d}'.format(fyear,fmonth)
                elif usedatedir == 'Yj':
                    datedir = '{}/{:03d}'.format(fyear,dirdoy)
                if debug:
                    print ("date related subdirectory:", datedir)
            path = os.path.normpath(f)
            li = path.split(os.sep)
            if not sensorid and not protocol in ['ftp','FTP']:
                if len(li) >= 2:
                    sensid = li[-2]
                if sensid == 'raw' and len(li) >= 3:  # in case an archive raw data structure is loaded
                    sensid = li[-3]
            elif not sensorid and protocol in ['ftp','FTP']:
                sensid = f.split('.')[0].rpartition('_')[0]
            else:
                sensid = sensorid

            destpath = createdestinationpath(destination, stationid, sensid, datedir=datedir, forcelocal=forcelocal)
            if debug:
                print ("Destinationpath:", destpath)

            destname = os.path.join(destpath,li[-1])

            if not os.path.isdir(destpath):
                os.makedirs(destpath)
            if debug:
                print ("   -> write destination (for raw files): {} , {}".format(destpath, li[-1]))

            if protocol in ['ftp','FTP']:
                fhandle = open(destname, 'wb')
                ftp.retrbinary('RETR ' + f, fhandle.write)
                fhandle.close()
                if deleteremote in [True,'True']:
                    ftp.delete(f)
            elif protocol in ['scp','SCP']:
                mm.scptransfer(user+'@'+address+':'+f,destpath,password,timeout=600)
            elif protocol in ['rsync']:
                # create a command line string with rsync ### please note,,, rsync requires password less comminuctaion
                if deleteremote in [True,'True']:
                    deleteopt = " --remove-source-files "
                else:
                    deleteopt = " "
                rsyncstring = "rsync -avz -e ssh{}{} {}".format(deleteopt, user+'@'+address+':'+f,destpath)
                print ("Executing:", rsyncstring)
                subprocess.call(rsyncstring.split())
            elif protocol in ['html','HTML']:
                pass
            elif protocol in ['']:
                if not os.path.exists(destname) or not filecmp.cmp(f, destname):
                    copyfile(f, destname)
                    if deleteremote in [True,'True']:
                        os.remove(f)
                else:
                    print ("   -> raw file already existing - skipping write")
            if zipping:
                if debug:
                    print (" raw data wil be zipped")
                dirname = os.path.dirname(destname)
                oldname = os.path.basename(destname)
                pname = os.path.splitext(oldname)
                if not pname[1] in [".zip",".gz",".ZIP",".GZ"]:
                    zipname = pname[0]+'.zip'
                    with zipfile.ZipFile(os.path.join(dirname,zipname), 'w') as myzip:
                        myzip.write(destname,oldname, zipfile.ZIP_DEFLATED)
                    os.remove(destname)
                    destname = os.path.join(dirname,zipname)
                else:
                    if debug:
                        print (" data is zipped already")
            localpathlist.append(destname)

        if protocol in ['ftp','FTP']:
            ftp.close()
    else:
        localpathlist = [elem for elem in filelist]

    if debug:
        print ("   => all files are now on local system: {}".format(localpathlist))

    return localpathlist



def write_data(config=None,localpathlist=None,debug=False):
    """
    DESCRIPTION
        Read local data and write to database  

    RETURNS
    """
    success = True
    if not config:
        config={}
    if not localpathlist:
        localpathlist=[]
    print("  Writing data to database and/or archive")

    db = config.get('db')
    stationid = config.get('stationid','')
    sensorid = config.get('sensorid','')
    force = config.get('forcerevision','')
    subdirectory = config.get('writedirectory','')
    writemode = config.get('writemode','replace')
    if not writemode in ['replace','overwrite']:
        # replace will replace existing data and leave the rest unchanged
        # overwrite will delete the file and write a new one
        writemode = 'replace'


    for f in localpathlist:
        data = DataStream()

        try:
            data = read(f)
        except:
            data = DataStream()

        if data.length()[0] > 0:
            if debug:
                print (" Dealing with {}. Length = {}".format(f,data.length()[0]))
                print (" -------------------------------")
                print (" data looks like: ", data.ndarray)
                #print ("SensorID in file: {}".format(data.header.get('SensorID')))

            # Station ID provided?
            statiddata = data.header.get('StationID','')

            if not stationid == '':
                if not statiddata == stationid and not statiddata == '':
                    print("   StationID's from file and provided one (or dir) are different!")
                    print ("   Using provided value")
                data.header['StationID'] = stationid.upper()
            else:
                if data.header.get('StationID','') == '':
                    print("   Could not find station ID in datafile")
                    print("   Please provide by using -t stationid")
                    #sys.exit()
                    success = False
                    continue

            if debug:
                print("  -> Using StationID", data.header.get('StationID'))

            # Sensor ID extractable?
            sensiddata = data.header.get('SensorID','')

            if not sensorid == '':
                if not sensiddata == sensorid and not sensiddata == '':
                    print("   SensorID's from file and provided one (or dir) are different!")
                    print ("   Using provided value")
                data.header['SensorID'] = sensorid
            else:
                if data.header.get('SensorID','') == '':
                    print("   Could not find sensor ID in datafile")
                    print("   Please provide in configuration or by using -s sensorid")
                    # Abort try clause
                    success = False
                    continue

            fixsensorid = data.header.get('SensorID')

            if debug:
                print("  -> Using SensorID", data.header.get('SensorID'))

            # Conversions
            if config.get(fixsensorid):
                comments = []
                offdict = {}
                comment = data.header.get('DataComments')
                if comment:
                    comments.append(comment)
                print ("  Found modification parameters - applying ...")
                paradict = config.get(fixsensorid)
                if paradict.get('defaulttimecolumn') == 'sectime':
                    print ("    -> secondary time column of raw data will be used as primary")
                    data = data.use_sectime()
                    comm1 = "secondary time colum moved to primary"
                    comments.append(comm1)
                keylist = data._get_key_headers()
                keylist.extend(['time','sectime'])
                for key in keylist:
                    offset = paradict.get(key)
                    #if debug:
                    #    print ("     offset for key {}: {}".format(key, offset))
                    if offset:
                        #try:
                        offset = float(offset)
                        if offset and key in ['time','sectime']:
                            offdict[key] = timedelta(seconds=offset)
                        elif offset:
                            offdict[key] = offset
                        comm = "applied an offset of {} to column {}".format(offset, key)
                        comments.append(comm)
                        print ("    -> offset of {} applied to {}".format(offset, key))
                        #except:
                        #    print ("    -> failure applying offset")
                if offdict:
                    if debug:
                        print ("    applying offsets: {}".format(offdict))
                    #'time': timedelta(hours=1), 'x': 4.2, 'f': -1.34242
                    data = data.offset(offdict)

                # extend comment
                if len(comments) > 1:
                    newcomment = ", ".join(comments) 
                    data.header['DataComments'] = newcomment
                elif len(comments) > 0:
                    newcomment = comments[0]
                    data.header['DataComments'] = newcomment
                print ("  => modifications done")

                #print ("DATAComments", comments, newcomment, data.header['DataComments'])

            def merge_two_dicts(x, y):
                z = x.copy()   # start with x's keys and values
                z.update(y)    # modifies z with y's keys and values & returns None
                return z

            datainfoid = data.header.get('DataID')
            if force:
                if not datainfoid:
                    datainfoid = "{}_{}".format(fixsensorid,str(force).zfill(4))
                    if debug:
                        print ("Using DataID: {}".format(datainfoid))

            # Get existing header information from database and combine with new info
            if datainfoid:
                existheader = db.fields_to_dict(datainfoid)
                data.header = merge_two_dicts(existheader,data.header)

            # Writing data
            if not debug and mm.get_bool(config.get('writedatabase')):
                print("  {}: Adding {} data points to DB now".format(data.header.get('SensorID'), data.length()[0]))

                if not len(data.ndarray[0]) > 0:
                    data = data.linestruct2ndarray()  # Dealing with very old formats                   
                if len(data.ndarray[0]) > 0:
                    try:
                        if not force == '':
                            tabname = "{}_{}".format(fixsensorid,str(force).zfill(4))
                            print (" - Force option chosen: forcing data to table {}".format(tabname))
                            print ("   IMPORTANT: general database meta information will not be updated")
                            db.write(data, tablename=tabname)
                        else:
                            db.write(data)
                    except:
                        print (" !! Error when writing ", f)
                        success = False
            elif debug:
                print ("  DEBUG selected - no database written")

            # Writing data
            if not debug and mm.get_bool(config.get('writearchive')):
                if force:
                    archivepath = os.path.join(config.get('archivepath'),stationid.upper(),fixsensorid,datainfoid)

                    if config.get('archiveformat'):
                        print("  {}: Writing {} data points archive".format(data.header.get('SensorID'), data.length()[0]))
                        if fixsensorid.startswith("LEMI"):
                            # LEMI bin files contains str1 column which cannot be written to PYCDF (TODO) - column contains GPS state
                            data = data._drop_column('str1')
                        if archivepath:
                            data.write(archivepath,filenamebegins=datainfoid+'_',format_type=config.get('archiveformat'),mode=writemode,subdirectory=subdirectory)
                else:
                    print ("  Writing to archive requires forcerevision")
            elif debug:
                print ("  DEBUG selected - no archive written")
    return success

def main(argv):
    version = "2.0.0"
    conf = ''
    etime = ''
    deptharg = ''
    writedbarg = ''
    writearchivearg = ''
    hostname = socket.gethostname().upper()
    statusmsg = {}
    debug = False
    filelist = []
    localpathlist = []
    proxies = {}

    try:
        opts, args = getopt.getopt(argv,"hc:e:d:w:a:D",["configuration=","endtime=","depth=","writedb=","writearchive=","debug=",])
    except getopt.GetoptError:
        print('file_download.py -c <configuration> -e <endtime> -d <depth>')
        sys.exit(2)
    for opt, arg in opts:
        if opt == '-h':
            print('-------------------------------------')
            print('Description:')
            print('file_download.py reads data from various sources ')
            print('and uploads data to a data bank.')
            print('Filtering and archiving is done using "cleanup".')
            print('-------------------------------------')
            print('Usage:')
            print('file_download.py -c <configuration> -e <endtime> -d <depth>')
            print('-------------------------------------')
            print('Options:')
            print('-c            : configuration file')
            print('-e            : endtime')
            print('-d            : depth: 1 means today, 2 today and yesterday, 3 last three days, etc')
            print('-w            : write to database')
            print('-a            : write to archive')
            print('-------------------------------------')
            print('Examples:')
            print('---------')
            print('---------')
            sys.exit()
        elif opt in ("-c", "--configuration"):
            conf = arg
        elif opt in ("-e", "--endtime"):
            etime = arg
        elif opt in ("-d", "--depth"):
            try:
                deptharg = int(arg)
                if not deptharg >= 1:
                    print("provided depth needs to be positive")
                    sys.exit()
            except:
                print("depth needs to be an integer") 
                sys.exit()
        elif opt in ("-w", "--writedb"):
            writedbarg = arg
        elif opt in ("-a", "--writearchive"):
            writearchivearg = arg
        elif opt in ("-D", "--debug"):
            debug = True

    # Read configuration file
    # -----------------------
    print ("Running file_download - version {}".format(version))
    print ("-------------------------------")

    name = "{}-file-download-{}".format(hostname, os.path.split(conf)[1].split('.')[0])
    statusmsg[name] = 'file_download successfully finished'

    if conf == '':
        print ('Specify a path to a configuration file using the  -c option:')
        print ('-- check archive.py -h for more options and requirements')
        sys.exit()
    else:
        if os.path.isfile(conf):
            print ("  Read file with GetConf")
            config = mm.get_conf(conf)
            print ("   -> configuration data extracted")
        else:
            print ('Specify a valid path to a configuration file using the  -c option:')
            print ('-- check archive.py -h for more options and requirements')
            sys.exit()

    if etime == '':
        current = datetime.now(timezone.utc).replace(tzinfo=None) # make that a variable
    else:
        current = testtime(etime)

    # check configuration information
    # -----------------------
    config, success = check_configuration(config=config, debug=debug)

    if not success:
        statusmsg[name] = 'invalid configuration data - aborting'
    else:
        # Override config data with given inputs
        # -----------------------
        if writearchivearg:
            config['writearchive'] = mm.get_bool(writearchivearg)
        if writedbarg:
            config['writedatabase'] = mm.get_bool(writedbarg)
        if deptharg:
            config['defaultdepth'] = deptharg

        # Create datelist
        # -----------------------
        datelist = get_datelist(config=config,current=current,debug=debug)

        # Obtain list of files to be transferred
        # -----------------------
        try:
            filelist = create_transfer_list(config=config,datelist=datelist,debug=debug)
            moveon = True
        except:
            statusmsg[name] = 'could not obtain remote file list - aborting'
            moveon = False

        if moveon:
            # Obtain list of files to be transferred
            # -----------------------
            try:
                localpathlist = obtain_data_files(config=config,filelist=filelist,debug=debug)
            except:
                statusmsg[name] = 'getting local file list failed - check permission'
                localpathlist = []

            # Write data to specified destinations
            # -----------------------
            #try:
            if config.get('db') and len(localpathlist) > 0 and (mm.get_bool(config.get('writedatabase')) or mm.get_bool(config.get('writearchive'))):
                succ = write_data(config=config,localpathlist=localpathlist,debug=debug)
                if not succ:
                    statusmsg[name] = 'could not write data - check logfile'
            #except:
            #    statusmsg[name] = 'problem when writing data'

    # Send Logging
    # -----------------------
    receiver = config.get('notification')
    receiverconf = config.get('notificationconf')
    logpath = config.get('logpath')

    if config.get('https'):
        proxies['https'] = config.get('https')
    if config.get('http'):
        proxies['http'] = config.get('http')

    if debug:   #No update of statusmessages if only a selected sensor list is analyzed
        print (statusmsg)
    else:
        martaslog = ml(logfile=logpath,receiver=receiver,proxies=proxies)
        martaslog.telegram['config'] = receiverconf
        martaslog.msg(statusmsg)

    print ("----------------------------------------------------------------")
    print ("file_download finished")
    print ("----------------------------------------------------------------")
    if statusmsg[name] == 'file_download successfully finished':
        print ("SUCCESS")
    else:
        print ("FAILURE")

if __name__ == "__main__":
   main(sys.argv[1:])





from __future__ import print_function
from __future__ import absolute_import

"""
Protocol for PalmAcq and ObsDAQ by MINGEO, Hungary

works for ObsDaqs since 55Fxxx serial numbers when connected to a
PalmAcq by a combined port cable
The MARTAS host is connected to PalmAcq by USB cable

Settings are made in a config file defined in martas.cfg like
obsdaqconfpath  :  /etc/martas/obsdaq.cfg

"""


# ###################################################################
# Import packages
# ###################################################################

import re     # for interpretation of lines
import struct # for binary representation
import socket # for hostname identification
import string # for ascii selection
from datetime import datetime, timezone
from twisted.protocols.basic import LineReceiver
from twisted.python import log
import math
import serial # for initializing command
import os,sys

# Relative import of core methods as long as martas is not configured as package
# import support for GetConf2 and dataToFile (can also be obtained from magpy)
from martas.core import methods as mm



def getRoundingFactor(lsb):
    # rounding factor is used for rounding reasonably
    return 10**(-1*round(math.log(lsb)/math.log(10)-1))


## Mingeo ObsDAQ protocol
##
class obsdaqProtocol(LineReceiver):
    """
    The Obsdaq protocol gets data assuming:
        connected to a PalmAcq
        PalmAcq is in Transparent mode (see manual)

    SETUP:
        1.) use palmacq.py to make settings of PalmAcq 
        2.) use obsdaq.py to make settings of ObsDAQ
    """
    def __init__(self, client, sensordict, confdict):
        """
        'client' could be used to switch between different publishing protocols
                 (e.g. MQTT or WS-MCU gateway factory) to publish events
        'sensordict' contains a dictionary with all sensor relevant data (sensors.cfg)
        'confdict' contains a dictionary with general configuration parameters (martas.cfg)
        """
        self.client = client
        self.sensordict = sensordict    
        self.confdict = confdict
        self.count = 0  ## counter for sending header information
        self.sensor = sensordict.get('sensorid')
        self.hostname = socket.gethostname()
        self.printable = set(string.printable)
        #log.msg("  -> Sensor: {}".format(self.sensor))
        self.datalst = []
        self.datalstSup = []
        self.datacnt = 0
        self.metacnt = 10
        self.errorcnt = {'time':0}

        self.delaylist = []  # delaylist contains up to 1000 diffs between gps and ntp
                             # the median of this values is used for ntp timedelay
        self.timedelay = 0.0
        self.timethreshold = 3 # secs - waring if timedifference is larger the 3 seconds

        # Serial configuration
        self.baudrate=int(sensordict.get('baudrate'))
        self.port = confdict['serialport']+sensordict.get('port')
        self.parity=sensordict.get('parity')
        self.bytesize=int(sensordict.get('bytesize'))
        self.stopbits=int(sensordict.get('stopbits'))
        if sys.version_info >= (3, 0):
            self.delimiter='\r'.encode('ascii')
        else:
            self.delimiter='\r'
        self.timeout=2 # should be rate dependend
        sensorSuplst = self.sensor.split('_')
        if len(sensorSuplst) == 3:
            self.sensoridSup = '_'.join([sensorSuplst[0]+'sup',sensorSuplst[1],sensorSuplst[2]])


        # QOS
        self.qos=int(confdict.get('mqttqos',0))
        if not self.qos in [0,1,2]:
            self.qos = 0
        log.msg("  -> setting QOS:", self.qos)

        # Debug mode
        debugtest = confdict.get('debug')
        self.debug = False
        if debugtest == 'True':
            log.msg('DEBUG - {}: Debug mode activated.'.format(self.sensordict.get('protocol')))
            self.debug = True    # prints many test messages
        else:
            log.msg('  -> Debug mode = {}'.format(debugtest))

        # get obsdaq specific constants
        self.obsdaqconf = mm.get_conf(self.confdict.get('obsdaqconfpath'))
        
        self.headernames = '[{},{},{},{}]'.format(self.obsdaqconf.get('NAME_X'),self.obsdaqconf.get('NAME_Y'),self.obsdaqconf.get('NAME_Z'),'ntptime')
        self.headerunits = '[{},{},{},{}]'.format(self.obsdaqconf.get('UNIT_X'),self.obsdaqconf.get('UNIT_Y'),self.obsdaqconf.get('UNIT_Z'),'none')
        CCdic = {'02':10.,'03':5.,'04':2.5}
        self.gainmax = CCdic[str(self.obsdaqconf.get('CC')).zfill(2)]
        self.scale_x = float(self.obsdaqconf.get('SCALE_X'))
        self.scale_y = float(self.obsdaqconf.get('SCALE_Y'))
        self.scale_z = float(self.obsdaqconf.get('SCALE_Z'))
        # least significant bit, smallest discrete value in given unit
        self.lsb_x = 2**-23 * self.gainmax * self.scale_x
        self.lsb_y = 2**-23 * self.gainmax * self.scale_y
        self.lsb_z = 2**-23 * self.gainmax * self.scale_z
        # rounding factor is used for rounding reasonably
        self.rfactor_x = getRoundingFactor(self.lsb_x)
        self.rfactor_y = getRoundingFactor(self.lsb_y)
        self.rfactor_z = getRoundingFactor(self.lsb_z)
        # header factor is used for sending decimal numbers as integer (or long)
        self.factor_x = 1
        if self.rfactor_x > 1.:
            self.factor_x = int(self.rfactor_x)
        self.factor_y = 1
        if self.rfactor_y > 1.:
            self.factor_y = int(self.rfactor_y)
        self.factor_z = 1
        if self.rfactor_z > 1.:
            self.factor_z = int(self.rfactor_z)
        self.headerfactors = '[{},{},{},{}]'.format(self.factor_x,self.factor_y,self.factor_z,1)

        # get constants for Obsdaq's supplementary channels
        self.headernamesSup = '[{},{},{},{},{}]'.format(self.obsdaqconf.get('NAME_V'),self.obsdaqconf.get('NAME_T'),self.obsdaqconf.get('NAME_P'),self.obsdaqconf.get('NAME_Q'),self.obsdaqconf.get('NAME_R'))
        self.headerunitsSup = '[{},{},{},{},{}]'.format(self.obsdaqconf.get('UNIT_V'),self.obsdaqconf.get('UNIT_T'),self.obsdaqconf.get('UNIT_P'),self.obsdaqconf.get('UNIT_Q'),self.obsdaqconf.get('UNIT_R'))
        self.scale_p = float(self.obsdaqconf.get('SCALE_P'))
        self.scale_q = float(self.obsdaqconf.get('SCALE_Q'))
        self.scale_r = float(self.obsdaqconf.get('SCALE_R'))
        # least significant bit, smallest discrete value in given unit = 1V / 8000 * scale , see manual
        self.lsb_p = 1./8000 * self.scale_p
        self.lsb_q = 1./8000 * self.scale_q
        self.lsb_r = 1./8000 * self.scale_r
        # rounding factor is used for rounding reasonably
        self.rfactor_p = getRoundingFactor(self.lsb_p)
        self.rfactor_q = getRoundingFactor(self.lsb_q)
        self.rfactor_r = getRoundingFactor(self.lsb_r)
        # header factor is used for sending decimal numbers as integer (or long)
        self.factor_p = 1
        if self.rfactor_p > 1.:
            self.factor_p = int(self.rfactor_p)
        self.factor_q = 1
        if self.rfactor_q > 1.:
            self.factor_q = int(self.rfactor_q)
        self.factor_r = 1
        if self.rfactor_r > 1.:
            self.factor_r = int(self.rfactor_r)
        # Obsdaq's voltage and temperature are measured and calculated fixed, see manual
        self.headerfactorsSup = '[{},{},{},{},{}]'.format('10000','1000',self.factor_p,self.factor_q,self.factor_r)

        # get constants for Obsdaq's supplementary channels
        self.headernamesSup = '[{},{},{},{},{}]'.format(self.obsdaqconf.get('NAME_V'),self.obsdaqconf.get('NAME_T'),self.obsdaqconf.get('NAME_P'),self.obsdaqconf.get('NAME_Q'),self.obsdaqconf.get('NAME_R'))
        self.headerunitsSup = '[{},{},{},{},{}]'.format(self.obsdaqconf.get('UNIT_V'),self.obsdaqconf.get('UNIT_T'),self.obsdaqconf.get('UNIT_P'),self.obsdaqconf.get('UNIT_Q'),self.obsdaqconf.get('UNIT_R'))


    def connectionMade(self):
        log.msg('  -> {} connected.'.format(self.sensor))

    def connectionLost(self, reason):
        log.msg('  -> {} lost.'.format(self.sensor))
        log.msg(reason)

    def processData(self, data):
        currenttime = datetime.now(timezone.utc).replace(tzinfo=None)
        outdate = datetime.strftime(currenttime, "%Y-%m-%d")
        filename = outdate
        sensorid = self.sensor
        datearray = []
        dontsavedata = False
        triggerflag = 0
        voltage = 0
        temp = 0
        x,y,z,p,q = 0,0,0,0,0

        packcode = '6hLlll6hL'
        header = "# MagPyBin %s %s %s %s %s %s %d" % (self.sensor, '[x,y,z,sectime]', self.headernames, self.headerunits, self.headerfactors, packcode, struct.calcsize(packcode))
        supplement = False
        packcodeSup = '6hLlllll'
        
        headerSup = "# MagPyBin %s %s %s %s %s %s %d" % (self.sensoridSup, '[var1,var2,var3,var4,var5]', self.headernamesSup, self.headerunitsSup, self.headerfactorsSup, packcode, struct.calcsize(packcode))

        if data.startswith(':R'):
            # :R,00,YYMMDD.hhmmss.sss,*xxxxxxyyyyyyzzzzzzt
            # :R,00,YYMMDD.hhmmss.sss,*xxxxxxyyyyyyzzzzzzt:vvvvttttppppqqqqrrrr
            # :R,00,200131.143739.617,*0259FEFFF1BFFFFCEDL:04AC11CC000B000B000B
            d = data.split(',')
            Y = int('20'+d[2][0:2])
            M = int(d[2][2:4])
            D = int(d[2][4:6])
            h = int(d[2][7:9])
            m = int(d[2][9:11])
            s = int(d[2][11:13])
            us = int(d[2][14:17]) * 1000
            try:
                timestamp = datetime(Y,M,D,h,m,s,us)
            except:
                # TODO properly update palmacq's GPS module
                # leap years until 2100(! :-)
                if M == 2:
                    if Y%4 and D>28:
                        D = D - 28
                        M = 3
                    elif not Y%4 and D>29:
                        D = D - 29
                        M = 3
                if D > 31:
                    # 2022-03-32 happened on 2022-04-01
                    D = D - 31
                    M = M+1
                if D == 31:
                    D = 1
                    M = M+1
                try:
                    timestamp = datetime(Y,M,D,h,m,s,us)
                except:
                    # if gps time is still not recognized, use birthday of Carl Friedrich Gausz
                    # (otherwise the sample would be discarded and the whole line is printed to the log file)
                    timestamp = datetime(1777,4,30)

            if d[3][0] == '*':
                x = (int('0x'+d[3][1:7],16) ^ 0x800000) - 0x800000
                # old line:
                #x = float(x) * 2**-23 * self.gainmax * self.scale_x
                x = round( float(x) * self.lsb_x * self.rfactor_x) / self.rfactor_x
                y = (int('0x'+d[3][7:13],16) ^ 0x800000) - 0x800000
                y = round( float(y) * self.lsb_y * self.rfactor_y) / self.rfactor_y
                z = (int('0x'+d[3][13:19],16) ^ 0x800000) - 0x800000
                z = round( float(z) * self.lsb_z * self.rfactor_z) / self.rfactor_z
                # (triggerflag not used here)
                triggerflag = d[3][19]
            else:
                typ = "none"
            sup = d[3].split(':')
            if len(sup) == 2:
                supplement = True
                voltage = int(sup[1][0:4],16) ^ 0x8000 - 0x8000
                voltage = float(voltage) * 2.6622e-3 + 9.15
                voltage = round(voltage*10000)/10000.
                temp = int(sup[1][4:8],16) ^ 0x8000 - 0x8000
                temp = float(temp) / 128.
                temp = round(temp*1000)/1000.
                p = (int('0x'+sup[1][8:12],16) ^ 0x8000) - 0x8000
                p = float(p) / 8000.0 * self.scale_p + float(self.obsdaqconf.get('DIFF_P'))
                p = round(p*self.rfactor_p)/self.rfactor_p
                q = (int('0x'+sup[1][12:16],16) ^ 0x8000) - 0x8000
                q = float(q) / 8000.0 * self.scale_q + float(self.obsdaqconf.get('DIFF_Q'))
                q = round(q*self.rfactor_q)/self.rfactor_q
                r = (int('0x'+sup[1][16:20],16) ^ 0x8000) - 0x8000
                r = float(r) / 8000.0 * self.scale_r + float(self.obsdaqconf.get('DIFF_R'))
                r = round(r*self.rfactor_r)/self.rfactor_r
            if self.debug:
                log.msg(str(timestamp)+'\t',end='')
                log.msg(str(x)+'\t',end='')
                log.msg(str(y)+'\t',end='')
                log.msg(str(z)+'\t',end='')
                log.msg(str(triggerflag))
                if len(sup) == 2:
                    log.msg('supplementary:\t',end='')
                    log.msg(str(voltage)+' V\t',end='')
                    log.msg(str(temp)+' degC\t',end='')
                    log.msg(str(p)+'\t',end='')
                    log.msg(str(q)+'\t',end='')
                    log.msg(str(r)+'\t')
            typ = "valid"
        else:
            typ = "none"
            if self.debug:
                log.msg(':R not found')
 
        if not typ == "valid": 
            dontsavedata = True

        if not typ == "none":
            datearray = mm.datetime_to_array(timestamp)
            try:
                datearray.append(int(x * self.factor_x))
                datearray.append(int(y * self.factor_y))
                datearray.append(int(z * self.factor_z))
                # add secondary time (NTP-time)
                datearray.extend(mm.datetime_to_array(currenttime))
                data_bin = struct.pack('<'+packcode,*datearray)
            except:
                data_bin = None
                log.msg('{} protocol: Error while packing binary data'.format(self.sensordict.get('protocol')))
            if not self.confdict.get('bufferdirectory','') == '':
                mm.data_to_file(self.confdict.get('bufferdirectory'), sensorid, filename, data_bin, header)
            returndata = ','.join(list(map(str,datearray)))

            if supplement:
                try:
                    datearraySup = mm.datetime_to_array(timestamp)
                    datearraySup.append(int(voltage *10000))
                    datearraySup.append(int(temp *1000))
                    datearraySup.append(int(p *self.factor_p))
                    datearraySup.append(int(q *self.factor_q))
                    datearraySup.append(int(r *self.factor_r))
                    data_bin_Sup = struct.pack('<'+packcodeSup,*datearraySup)
                except:
                    log.msg('{} protocol: Error while packing binary supplement data'.format(self.sensordict.get('protocol')))
                if not self.confdict.get('bufferdirectory','') == '':
                    mm.data_to_file(self.confdict.get('bufferdirectory'), self.sensoridSup, filename, data_bin_Sup, headerSup)
                returndataSup = ','.join(list(map(str,datearraySup)))
            else:
                returndataSup = ''
                headerSup = ''

        else:
            returndata = ''
            returndataSup = ''
            headerSup = ''

        return returndata, header, returndataSup, headerSup

         
    def lineReceived(self, line):
        topic = self.confdict.get('station') + '/' + self.sensordict.get('sensorid')
        topicSup = self.confdict.get('station') + '/' + self.sensoridSup
        if sys.version_info >= (3, 0):
            line = line.decode('ascii')
        # extract only ascii characters 
        line = ''.join(filter(lambda x: x in string.printable, line))
        ok = True
        try:
            data, head, dataSup, headSup = self.processData(line)
        except:
            print('{}: Data seems not to be PalmAcq data: Looks like {}'.format(self.sensordict.get('protocol'),line))
            ok = False

        if ok:
            senddata = False
            coll = int(self.sensordict.get('stack'))
            if coll > 1:
                self.metacnt = 1 # send meta data with every block
                if self.datacnt < coll:
                    self.datalst.append(data)
                    if dataSup:
                        self.datalstSup.append(dataSup)
                    self.datacnt += 1
                else:
                    senddata = True
                    data = ';'.join(self.datalst)
                    if len(self.datalstSup)>0:
                        dataSup = ';'.join(self.datalstSup)
                    self.datalst = []
                    self.datalstSup = []
                    self.datacnt = 0
            else:
                senddata = True

            if senddata:
                self.client.publish(topic+"/data", data, qos=self.qos)
                if dataSup:
                    self.client.publish(topicSup+"/data", dataSup, qos=self.qos)
                if self.count == 0:
                    add = "SensorID:{},StationID:{},DataPier:{},SensorModule:{},SensorGroup:{},SensorDecription:{},DataTimeProtocol:{}".format( self.sensordict.get('sensorid',''),self.confdict.get('station',''),self.sensordict.get('pierid',''),self.sensordict.get('protocol',''),self.sensordict.get('sensorgroup',''),self.sensordict.get('sensordesc',''),self.sensordict.get('ptime','') )
                    self.client.publish(topic+"/dict", add, qos=self.qos)
                    self.client.publish(topic+"/meta", head, qos=self.qos)
                    if headSup:
                        self.client.publish(topicSup+"/meta", headSup, qos=self.qos)
                self.count += 1
                if self.count >= self.metacnt:
                    self.count = 0




#!/usr/bin/env python

import sys, time, os, socket
import serial
import struct, binascii, re, csv
from datetime import datetime, timedelta
from matplotlib.dates import date2num, num2date
import numpy as np
import time

port = '/dev/ttyS0'
baudrate='115200'
eol = '\r'

def lineread(ser,eol):
            # FUNCTION 'LINEREAD'
            # Does the same as readline(), but does not require a standard 
            # linebreak character ('\r' in hex) to know when a line ends.
            # Variable 'eol' determines the end-of-line char: '\x00'
            # for the POS-1 magnetometer, '\r' for the envir. sensor.
            # (Note: required for POS-1 because readline() cannot detect    
            # a linebreak and reads a never-ending line.)
            ser_str = ''
            timeout = time.time()+2
            while True:
                char = ser.read()
                if char == eol:
                    break
                if time.time() > timeout:
                    break
                ser_str += char
            return ser_str

def send_command(ser,command,eol,hex=False):

    command = eol+command+eol
    #print 'Command:  %s \n ' % command.replace(eol,'')
    sendtime = date2num(datetime.utcnow())
    #print "Sending"
    ser.write(command)
    #print "Received something - interpretation"
    response = lineread(ser,eol)
    #print "interprete"
    receivetime = date2num(datetime.utcnow())
    meantime = np.mean([receivetime,sendtime])
    #print "Timediff", (receivetime-sendtime)*3600*24
    return response, num2date(meantime).replace(tzinfo=None)

ser = serial.Serial(port, baudrate=baudrate , parity='N', bytesize=8, stopbits=1, timeout=2)
for i in range(99):
    call = str(i).zfill(2)+'TR00002'
    print(call)
    answer, actime = send_command(ser,call,eol)
    print(answer)




#!/usr/bin/env python
"""
Small program to send commands to an Arduino.
Used to interact with MARTAS supported scripts on Arduinos.
"""

import sys, time, os, socket
import getopt
import serial
import struct, binascii, re, csv
from datetime import datetime, timedelta
from matplotlib.dates import date2num, num2date
import numpy as np
import time
# get core functions to extract sensorcfg
from martas.core.methods import martaslog as ml
from martas.core import methods as mm
from martas.version import __version__


def get_arduino_cfg(mpath="/etc/martas/martas.cfg",debug=False):
    if debug:
        print ("Getting arduino config from sensors.cfg")
    serialcfg = {"port":"/dev/ttyACM0","baudrate":9600,"parity":"N","bytesize":8,"stopbits":1}
    try:
        conf = mm.get_conf(mpath)
        sensors = mm.get_sensors(conf.get('sensorsconf'))
        for sensordict in sensors:
            if sensordict.get("sensorid").find("ARDUINO")>-1:
                serialcfg["baudrate"]=int(sensordict.get('baudrate'))
                serialcfg["port"]=conf['serialport']+sensordict.get('port')
                serialcfg["parity"]=sensordict.get('parity')
                serialcfg["bytesize"]=int(sensordict.get('bytesize'))
                serialcfg["stopbits"]=int(sensordict.get('stopbits'))
        if debug:
            print ("Obtained the following parameters: {}".format(serialcfg))
    except:
        if debug:
            print ("Failed to obtain parameters from sensor.cfg file")
    return serialcfg


def send_command(ser,command,eol,hex=False):

    #use printable here
    response = ''
    fullresponse = ''
    maxcnt = 50
    cnt = 0
    command = command+eol
    #if(ser.isOpen() == False):
    #    ser.open()
    ser.flush()
    if sys.version_info >= (3, 0):
        ser.write(command.encode('ascii'))
    else:
        ser.write(command)
    # skipping all empty lines
    while response == '':
        response = ser.readline()
        if sys.version_info >= (3, 0):
            response = response.decode()
    # read until end-of-messageblock signal is obtained (use some break value)
    while not response.startswith('<MARTASEND>') and not cnt == maxcnt:
        cnt += 1
        fullresponse += response
        response = ser.readline()
        if sys.version_info >= (3, 0):
            response = response.decode()
    if cnt == maxcnt:
        fullresponse = 'Maximum count {} was reached'.format(maxcnt)
    return fullresponse


def main(argv):
    port = ""  # default
    baudrate=None
    parity=""
    bytesize=None
    stopbits=None
    timeout=2   # 0 for non-blocking read, does not work
    command = 'Status'
    eol = '\r\n'
    martaspath="/etc/martas/martas.cfg"
    debug = False
    travistestrun = False

    usagestring = 'ardcomm.py -c <command> -p <port> -b <baudrate> -a <parity> -y <bytesize> -s <stopbits> -t <timeout> -e <eol> -m <martaspath>'
    try:
        opts, args = getopt.getopt(argv,"hc:p:b:a:y:s:te:m:DT",["command=","port=","baudrate=","parity=","bytesize=","stopbits=","timeout=","eol=","martaspath=","debug=","Test=",])
    except getopt.GetoptError:
        print ('Check your options:')
        print (usagestring)
        sys.exit(2)
    for opt, arg in opts:
        if opt == '-h':
            print ('------------------------------------------------------')
            print ('Usage:')
            print (usagestring)
            print ('------------------------------------------------------')
            print ('Options:')
            print ('-h                             help')
            print ('-c                             command to send:')
            print ('                               Status          : return the current switch status')
            print ('                               P:switch:pin  : e.g. swP:0:4 = switch off pin 4')
            print ('-p                             set port - default is /dev/ttyACM0')
            print ('-b                             baudrate - default is 9600')
            print ('-a                             parity - default is "N"')
            print ('-y                             bytesize - default is 8')
            print ('-s                             stopbits - default is 1')
            print ('-t                             timeout - default is 0')
            print ('-e                             end of line - default is /r/n')
            print ('-m                             martas.cfg location - default is /etc/martas/martas.cfg')
            print ('------------------------------------------------------')
            print ('Examples:')
            print ('1. Switch on pin 4')
            print ('   python ardcomm.py -c "swP:1:4')
            sys.exit()
        elif opt in ("-c", "--command"):
            command = arg
            n = command.count(":")
            miss=-(n-2)
            for n in range(miss):
                command += ':'
        elif opt in ("-a", "--parity"):
            parity = arg
        elif opt in ("-y", "--bytesize"):
            bytesize = arg
        elif opt in ("-e", "--eol"):
            eol = arg
        elif opt in ("-b", "--baudrate"):
            baudrate = arg
        elif opt in ("-p", "--port"):
            port = arg
        elif opt in ("-t", "--timeout"):
            try:
                timeout = int(arg)
            except:
                timeout = arg
        elif opt in ("-y", "--bytesize"):
            try:
                bytesize = int(arg)
            except:
                bytesize = arg
        elif opt in ("-s", "--stopbits"):
            try:
                stopbits = int(arg)
            except:
                stopbits = arg
        elif opt in ("-m", "--martaspath"):
            martaspath = arg
        elif opt in ("-D", "--debug"):
            debug = True
        elif opt in ("-T", "--Test"):
            travistestrun = True

    serialcfg = get_arduino_cfg(mpath=martaspath,debug=debug)
    if not baudrate:
         baudrate=serialcfg.get("baudrate",9600)
    if not port:
         port=serialcfg.get("port","/dev/ttyACM0")
    if not parity:
         parity=serialcfg.get("parity","N")
    if not bytesize:
         bytesize=serialcfg.get("bytesize",8)
    if not stopbits:
         stopbits=serialcfg.get("stopbits",1)

    ser = serial.Serial(port, baudrate=baudrate , parity=parity, bytesize=bytesize, stopbits=stopbits, timeout=timeout)
    time.sleep(2)

    # Additional configs:
    ser.writeTimeout = 2     #timeout for write or write_timeout
    ser.write_timeout = 2

    # Open Connection
    # try it 10 time, then close and reopen serial - likely not necessary
    count = 0
    portopened = False
    while count < 10:
        try:
            ser.open()
            count = 10
            if debug:
                print ("ardcomm: serial port opened")
            portopened = True
        except Exception as e:
            count += 1
            #print ("ardcomm: error open serial port {}: {}".format(count,str(e)))
            if count == 8:
                try:
                    #closing if already open, so that it is reopenning for last try...
                    ser.close()
                except:
                    pass

    if not portopened and travistestrun:
        print ("ardcomm: serial port not available in testrun - finishing")
        sys.exit(0)


    if ser.isOpen():
        #try:
        #if debug:
        print("Sending command: {} ...".format(command))
        answer = send_command(ser,command,eol)
        #if debug:
        print("Answer: {}".format(answer))
        ser.close()
        #except Exception as e1:
        #    print ("ardcomm: communication error: {}".format(e1))
    else:
        print ("ardcomm: cannot open serial port - aborting")
        sys.exit(0)

if __name__ == "__main__":
   main(sys.argv[1:])





#!/usr/bin/env python
# coding=utf-8


import sys
from os import WCONTINUED

sys.path.insert(1, '/home/leon/Software/magpy/')  # should be magpy2

import unittest
from martas.core import methods as mm
from martas.version import __version__
from martas.app.monitor import _latestfile
from magpy.stream import *
from magpy.core import database
from magpy.core import plot as mp
import re
import os
import subprocess
#import telepot
#from telepot.loop import MessageLoop
import sys, getopt
import glob
import psutil
import signal
from crontab import CronTab
from pathlib import Path
from dateutil.parser import parse


class ActionHandler(object):
    """
    DESCRIPTION
        Action handler contains a command dictionary and will perform actions according to this
         dictionary dependend on any received command.
         In order to generate specific options, the Action handler needs also access to data and
         configurations of MARTAS/MARCOS
    METHODS

| class           |    method          |  version |  tested  |              comment       | manual | *used by |
| --------------- |  ----------------- |  ------- |  ------- |  ------------------------- | ------ | ---------- |
|  **ActionHandler**  |                |          |          |                            |        | telegrambot |
|  ActionHandler  |  __init__          |  2.0.0   |      yes |                            | -      |          |
|  ActionHandler  |  _getcam           |  2.0.0   |     *yes |                            | -      |          |
|  ActionHandler  |  _getspcae         |  2.0.0   |     *yes | by action_status           | -      |          |
|  ActionHandler  |  _identify_sensors |  2.0.0   |     *yes | by action_getdata          | -      |          |
|  ActionHandler  |  _identify_dates   |  2.0.0   |     *yes | by action_getdata          | -      |          |
|  ActionHandler  |  _jobprocess       |  2.0.0   |     *yes | by action_status           | -      |          |
|  ActionHandler  |  _tail             |  2.0.0   |     *yes | by action_getlog           | -      |          |
|  ActionHandler  |  set_default_configuration | 2.0.0 | yes |                            | -      |          |
|  ActionHandler  |  set_default_commands | 2.0.0 |      yes |                            | -      |          |
|  ActionHandler  |  init_martas       |  2.0.0   |     *yes | by __init__                | -      |          |
|  ActionHandler  |  interpret         |  2.0.0   |      yes |                            | -      |          |
|  ActionHandler  |  execute_command   |  2.0.0   |      yes |                            | -      |          |
|  ActionHandler  |  action_help       |  2.0.0   |     *yes |                            | -      |          |
|  ActionHandler  |  action_hello      |  2.0.0   |     *yes |                            | -      |          |
|  ActionHandler  |  action_martas     |  2.0.0   |      yes |                            | -      |          |
|  ActionHandler  |  action_marcos     |  2.0.0   |      yes |                            | -      |          |
|  ActionHandler  |  action_cam        |  2.0.0   |     *yes |                            | -      |          |
|  ActionHandler  |  action_reboot     |  2.0.0   |     *yes |                            | -      |          |
|  ActionHandler  |  action_martasupgrade | 2.0.0  |     *yes |                            | -      |          |
|  ActionHandler  |  action_system     |  2.0.0   |     *yes |                            | -      |          |
|  ActionHandler  |  action_martas_sensors | 2.0.0 |    *yes |                            | -      |          |
|  ActionHandler  |  action_figure     |  2.0.0   |     *yes |                            | -      |          |
|  ActionHandler  |  action_status     |  2.0.0   |     *yes |                            | -      |          |
|  ActionHandler  |  action_getlog     |  2.0.0   |     *yes |                            | -      |          |
|  ActionHandler  |  action_getdata    |  2.0.0   |     *yes |                            | -      |          |
|  ActionHandler  |  action_upload     |  2.0.0   |     *yes |                            | -      |          |
|  ActionHandler  |  action_getip      |  2.0.0   |     *yes |                            | -      |          |
|  ActionHandler  |  action_tmate      |  2.0.0   |     *yes |                            | -      |          |
|  ActionHandler  |  action_plot       |  2.0.0   |     *yes |                            | -      |          |

* all action_ methods are tested by execute_command
    APPLICATION

    RETURN
        will create a dictionary with
        {call :  {'message' : {'text': xxx, 'pic...
    """

    def __init__(self, configpath=None, commands=None, debug=False):
        # set some general parameters
        # read a "interpretation" dictionary from a file
        if not commands:
            commands = self.set_default_commands()
        else:
            # TODO eventually add new commands and replace defaults
            pass
        self.debug = debug
        # Now combine defauls with constructors provided
        if configpath:
            if debug:
                 print ("TelegramBot: Reading configuration")
            config = mm.get_conf(configpath)
        else:
            config = self.set_default_configuration()

        if debug:
            print ("Configuration:")
            print (config)
        # PLEASE NOTE: CONFIG should replace or extend contents of default dictionaries, not the whole dic
        self.configuration = config
        self.commanddict = commands
        #self.logger = self.logger_setup(name=config.get('logname','telegrambot'),loglevel=config.get('loglevel','INFO'),path=config.get('logging','stdout'))
        self.inputcounter = 0
        self.quest = False
        self.cvals = {}
        # some cleanups for later
        self.configuration['purpose'] = self.configuration.get('purpose').upper()
        # parameters will be filled by data and configuartion specific parameters for the requested actions
        self.messageconfig = {}
        if 'MARTAS' in self.configuration.get('purpose'):
            self.init_martas(self.configuration.get('martasconfig',""))
        self.messageconfig['tmppath'] = "/tmp"


    def _getcam(self, camport):
        """
        DESCRIPTION
            obtain device call for cam port
        """
        po = re.search(r'\d+', camport).group()
        if po and int(po) < 10:
            camport = "/dev/video{}".format(po)
        return camport


    def _getspace(self):
        """
        DESCRIPTION
            get some memory information and process status reports
        """
        statvfs = os.statvfs('/home')
        total = (statvfs.f_frsize * statvfs.f_blocks / (1024.*1024.))     # Size of filesystem in bytes
        remain = (statvfs.f_frsize * statvfs.f_bavail / (1024.*1024.))     # Number of free bytes that ordinary users
        mesg = "MEMORY status:\n----------\nDisk-size: {:.0f}MB\nDisk available: {:.0f}MB\nDisk occupied: {:.1f}%".format(total,remain, 100-(remain/total*100.))
        try:
            mem = psutil.virtual_memory()
            cpu = psutil.cpu_percent(interval=1)
            avail = mem.available / (1024*1024)
            total = mem.total / (1024*1024)
            mesg += "\nMemory total: {}MB\nMemory available: {}MB\nCPU usage: {}%\n".format(total,avail,cpu)
        except:
            pass

        return mesg

    def _identify_sensor(self, text):
        """
        DESCRIPTION
            check whether the text contains valid sensorid and returns a list of them
        """
        sensorlist = self.messageconfig.get("sensorlist",[])

        senslist = []
        if not text:
            return []
        splittext = text.split()
        validsensors = [s.get('sensorid').replace('$', '').replace('?', '').replace('!', '') for s in sensorlist]
        for element in splittext:
            if element in validsensors:
                # found a sensorid in within the text
                senslist.append(element)
        validnames = list(set([s.get('name') for s in sensorlist]))
        for element in splittext:
            if element in validnames:
                # found a sensorname in the text
                correspondingids = [s.get('sensorid').replace('$', '').replace('?', '').replace('!', '') for s in
                                    sensorlist if s.get('name') == element]
                if len(correspondingids) > 0:
                    for el in correspondingids:
                        senslist.append(el)
        if len(senslist) > 0:  # remove duplicates
            senslist = list(set(senslist))
        return senslist

    def _identify_dates(self, text):
        """
        DESCRIPTION
            extract dates from a text
        """
        # For old parse date versions
        def hasNumbers(inputString):
            return any(char.isdigit() for char in inputString)

        if not hasNumbers(text):
            return None

        try:
            dt = parse(text, fuzzy=True)
            dt = dt.replace(tzinfo=None)
        except:
            dt = None
        return dt

    def _jobprocess(self, debug=False):
        """
        DESCRIPTION
            Identify PIDs of MARTAS and MARCOS processes
        """
        lines = []
        # Status of MARTAS MARCOS jobs
        mycron = CronTab(user=True)
        for job in mycron:
            if debug:
                print("Testing job:", job)
            comm = job.comment
            cand = job.command
            en = job.is_enabled()
            cl = cand.split()
            if comm.find("MARCOS") >= 0 or comm.find("MARTAS") >= 0:
                if debug:
                    print (comm)
                active = "inactive"
                k = comm.replace("Running MARCOS process ", "")
                pidname = Path(cl[2]).stem
                if debug:
                    print("Testing pid", pidname)
                p = self.get_pid(pidname)
                if debug:
                    print (" - got pid", p)
                if p > 0:
                    active = "active"
                line = "{} : {}\n----------".format(comm, active)
                lines.append(line)

        mesg = "\n".join(lines)
        return mesg


    def _tail(self, path, lines=1, _buffer=4098):
        """
        DESCRIPTION:
            Obtain the last n line of a file f
            see: https://stackoverflow.com/questions/136168/get-last-n-lines-of-a-file-similar-to-tail
        """
        lines_found = []
        mesg = ""

        # block counter will be multiplied by buffer
        # to get the block size from the end
        block_counter = -1

        with open(path, "rb") as f:
            # loop until we find X lines
            while len(lines_found) < lines:
                try:
                    f.seek(block_counter * _buffer, os.SEEK_END)
                except IOError:  # either file is too small, or too many lines requested
                    f.seek(0)
                    lines_found = f.readlines()
                    break
                lines_found = f.readlines()
                # we found enough lines, get out
                # Removed this line because it was redundant the while will catch
                # it, I left it for history
                # if len(lines_found) > lines:
                #    break
                # decrement the block counter to get the
                # next X bytes
                block_counter -= 1

            mesg = "\n".join([x.decode() for x in lines_found[-lines:]])
        return mesg


    def set_default_configuration(self):
        """
        DESCRIPTION
            some defaults for configuration parameters
        """
        config = {}
        config['bot_id'] = ''
        config['allowed_users'] = ''
        config['purpose'] = 'MARTAS'
        config['base'] = '/home/leon/.martas'
        config['martasapp'] = '/home/leon/.martas/app'
        config['martaslog'] = '/home/leon/.martas/log'
        config['martasconfig'] = '/home/leon/.martas/conf/martas.cfg'
        config['marcosconfig'] = '/home/leon/.martas/conf/marcos.cfg'
        config['dbcredentials'] =  'cobsdb'
        config['proxy'] = None
        config['proxyport'] = None
        config['currentdatapath'] =  '/srv/products/data/current.data'
        config['defaultplot'] =  "/tmp/martas-demo.jpg"
        config['fig1'] =  "/tmp/Spectra.png"
        config['outlier'] = {"threshold" : 4}
        config['imbotoverview'] =  '/home/leon/.imbot/quickreport.py'
        config['imbotmemory'] =  '/second'
        config['imbotarchive'] =  '/srv/imbot'
        config['tmppath'] =  "/tmp"
        config['camport'] =  None
        config['camoptions'] = "-r 1280x720 -S 100 -F 1 -D 5 --no-banner"
        config['uploadmemory'] = '/home/leon/.martas/memory/uploadmemory.json'
        config['uploadconfig'] = ''
        config['bot_logging'] =  '/home/leon/.martas/log/telegrambot.log'
        config['loglevel'] =  'INFO'

        # Move those to commands dict
        config['figure1'] =  ["spectrum", "spectra"]
        config['cam'] =  ["snapshot", "picture", "cheese"]
        config['getlog'] =  ["getthelog"]

        # NOT IN telegrambot.cfg
        config['logname'] = 'telegrambot'
        config['logging'] = 'stdout'
        config['travistestrun'] = False
        return config

    def set_default_commands(self):
        """
        DESCRIPTION
            some defaults for configuration parameters
        """
        
        commanddict = {'hello' :  {'commands': ['hello','Hello'],
                                   'combination' : 'any',
                                   'priority' : 1,
                                   'method' : 'hello',
                                   'availability': ['all'],
                                   'description': 'say hello, bot'},
                       'help'  :  {'commands': ['help','Help'],
                                   'combination' : 'any',
                                   'priority' : 1,
                                   'availability': ['all'],
                                   'description': 'print descriptions'},
                       'sensor' : {'commands': ['sensors','sensor','Sensors','Sensor'],
                                   'combination' : 'any',
                                   'priority' : 1,
                                   'options' : {'variable':True},
                                   'availability': ['MARTAS'],
                                   'description': 'get sensors from config and check whether recent buffer data are existing\n  Command options:\n  sensors\n  sensor sensorid or sensors sensorname (provides some details on the selected sensor)'},
                       'imbot' :  {'commands': ['imbot','IMBOT'],
                                   'availability': ['IMBOT'],
                                   'priority' : 1,
                                   'options' : {'variable':True},
                                   'combination' : 'any',
                                   'description': 'get some IMBOT status messages:'},
                       'system' : {'commands': ['System','system'], 
                                   'availability': ['all'],
                                   'combination' : 'any',
                                   'priority' : 1,
                                   'description': 'get some basic information an the remote system and its software (hardware, magpy version)'},
                       'cam' :    {'commands': ['cam','Cam','picture','Picture','photo'],
                                   'combination' : 'any',
                                   'priority' : 1,
                                   'options' : {'device':True},
                                   'availability': ['all'],
                                   'description': 'get a live picture from a connected camera'},
                       'status' : {'commands': ['Status','status','Memory','memory','disk','space','Disk'],
                                   'combination' : 'any',
                                   'priority' : 1,
                                   'availability': ['all'],
                                   'description': 'get information on disk space, memory, and martas-marcos processes'},
                       'getlog' : {'commands': ['getlog','get log','get the log', 'print log', 'print the log', ' log'],
                                   'combination' : 'any',
                                   'priority' : 1,
                                   'options' : {'logfile':True},
                                   'availability': ['all'],
                                   'description': 'obtain last n lines of a log file\n  Command options:\n  getlog  \n  getlog 10  (last 10 lines)  \n  getlog 10 syslog  (telegrambot, martas, syslog, messages)'},
                       'getdata': {'commands': ['data'],
                                   'combination' : 'any',
                                   'priority' : 1,
                                   'options' : {'variable':True},
                                   'availability': ['MARTAS'],
                                   'description': 'get sensor data\n Command options:\n  use datetime and sensorid\n  e.g. get data from 2020-11-22 11:22 of LEMI025_22_0003'},
                       'plot' :   {'commands': ['plot','Plot'],
                                   'combination' : 'any',
                                   'priority' : 1,
                                   'options' : {'variable':True},
                                   'availability': ['MARTAS'],
                                   'description': 'provide sensorid : get diagram of specific sensor by default of the last 24 h \n  Command options:\n  plot sensorid\n  plot sensorid starttime\n  plot sensorid starttime endtime'},
                       'switch' : {'commands': ['switch','Switch'],
                                   'combination' : 'any',
                                   'priority' : 1,
                                   'availability': ['MARTAS'],
                                   'options' : {'swP:0:4' : ['P:0:4','swP:0:4','heating off','pin4 off','off'], 'swP:1:4' : ['P:1:4','swP:1:4','heating on','pin4 on','on'], 'swP:1:5' : ['P:1:5','swP:1:5','pin5 on'], 'swP:0:5' : ['P:0:5','swP:0:5','pin5 on'], 'swD' : ['swD','state','State'] },
                                   'description': 'optional: turn on/off remote switches if supported by the hardware (work in progress). switch without options will return the switch status. switch swP:1:4  will turn on port 4 from a attached microcontroller'},
                       'badwords':{'commands': ['fuck','asshole'],
                                   'combination' : 'any',
                                   'priority' : 1,
                                   'availability': ['hidden'],
                                   'description': ''},
                       'figure1': {'commands': ['figure1','Figure1','fig1','Fig1'],
                                   'combination' : 'any',
                                   'priority' : 1,
                                   'availability': ['all'],
                                   'description': 'open a preconfigured figure'},
                       'figure2': {'commands': ['figure2','Figure2','fig2','Fig2'],
                                   'combination' : 'any',
                                   'priority' : 1,
                                   'availability': ['all'],
                                   'description': 'open a preconfigured figure'},
                       'reboot': {'commands': ['reboot'],
                                   'combination' : 'any',
                                   'priority' : 1,
                                   'availability': ['hidden'],
                                   'description': 'reboot the remote computer'},
                       'martasupgrade': {'commands': ['upgrade'],
                                   'combination' : 'any',
                                   'priority' : 1,
                                   'availability': ['hidden'],
                                   'description': 'upgrade MARTAS to its newest version'},
                       'version': {'commands': ['version'],
                                        'combination': 'any',
                                        'priority': 1,
                                        'availability': ['all'],
                                        'description': 'library version 2.0.0'},
                       'martas' : {'commands': ['Martas','martas','MARTAS'],
                                   'combination' : 'any',
                                   'availability': ['MARTAS'],
                                   'priority' : 2,
                                   'options' : {'restart':['restart'],'start':['start'],'stop':['stop'],'status':['status']},
                                   'description': 'restart-stop-start-update-status : e.g. restart MARTAS process: send - martas restart - to restart a martas process'},
                       'marcos' : {'commands': ['Marcos','marcos','MARCOS'],
                                   'combination' : 'any',
                                   'availability': ['MARCOS'],
                                   'priority' : 2,
                                   'options' : {'restart':['restart'],'start':['start'],'stop':['stop'],'status':['status']},
                                   'description': 'restart-stop-start-status : e.g. restart MARCOS processes: if you have a collector called collect-local then - marcos local restart - will restart this process'},
                       'aperta': {'commands': ['aperta', 'Aperta', 'Sesam öffne dich'],
                                        'combination': 'any',
                                        'priority': 1,
                                        'availability': ['hidden'],
                                        'description': 'Open secret door'},
                       'upload': {'commands': ['upload', 'send data', 'Upload', 'nach Hause telefonieren'],
                                        'combination': 'any',
                                        'priority': 1,
                                        'availability': ['hidden'],
                                        'description': 'Upload data to server'},
                       'getip': {'commands': ['getIP', ' IP ', 'IP ', 'getip', 'Getip', 'GetIP'],
                                        'combination': 'any',
                                        'priority': 1,
                                        'availability': ['hidden'],
                                        'description': 'send current IP address'},
                       }
        return commanddict


    def init_martas(self, martasconfig="", debug=False):
        """
        DESCRIPTION
            special initialization for martas machine
            -> reads martas config files (martas.cfg and sensors.cfg)
        APPLICATION:
            at the end of general init (after logger)
        """
        if not martasconfig:
            return
        conf = mm.get_conf(martasconfig)

        sensorlist = mm.get_sensors(conf.get('sensorsconf'))
        ardlist = mm.get_sensors(conf.get('sensorsconf'),identifier='?')
        sensorlist.extend(ardlist)
        owlist = mm.get_sensors(conf.get('sensorsconf'),identifier='!')
        sensorlist.extend(owlist)
        sqllist = mm.get_sensors(conf.get('sensorsconf'),identifier='$')
        sensorlist.extend(sqllist)
        self.messageconfig['mqttpath'] = conf.get('bufferdirectory','')
        self.messageconfig['sensorlist'] = sensorlist
        self.messageconfig['martaslog'] = conf.get('logging','')


    def interpret(self, input, exclude=None, debug=False):
        """
        DESCRIPTION
            Receive any text and identify existing commands and priorities within a command dictionary. Select the
            appropriate command(s) and return them as action dictionary.
        PARAMETERS:
            input : any input request
            exclude : a list of availability options to exclude (i.e. hidden)
        REQUIRES
            self.configuration
        RETURNS:
            a list of action items from commanddict in applictaion order
        """
        actiondict = {}
        if not exclude:
            exclude = []
        # Identify words, amount and positions of commandlist in input.
        for elem in self.commanddict:
            jobdic = self.commanddict.get(elem)
            jobname = elem
            jobpriority = jobdic.get('priority', 0)
            jobcommands = jobdic.get('commands', [jobname])
            availability = jobdic.get('availability', [])
            if debug:
                print ("Analysing {} with {}".format(elem, jobcommands))
            for jc in jobcommands:
                if jc in input and not any([x in exclude for x in availability]):
                    if debug:
                        print ("  -> found")
                    ad = actiondict.get(jobname,{})
                    ad['count'] = input.count(jc)
                    ad['priority'] = jobpriority
                    ad['position'] = input.find(jc)
                    actiondict[jobname] = ad

        actiondict = dict(sorted(actiondict.items(), key=lambda x: x[1]['position'], reverse=False))
        actiondict = dict(sorted(actiondict.items(), key=lambda x: x[1]['count'], reverse=True))
        actiondict = dict(sorted(actiondict.items(), key=lambda x: x[1]['priority'], reverse=False))
        actionlist = [el for el in actiondict]

        return actionlist

    def execute_action(self, actionlist, input="", debug=False):
        """
        DESCRIPTION
            Takes the action list and performs the requested actions. It will then recieve a result (message)
            dictionary from each action.
        PARAMETERS:
            actionlist : an actionlist as determined by interpret
            input : the original input to identify additional parameters for specific jobs
        RETURNS:
            message dictionary with results for each action
            messagedict : {'action' : {'message' : {'text' : "hello"},
                                                   {'pictures' : "hello"},
                                      }
        """
        messagedict = {}

        for action in actionlist:
            adict = {}
            msg = {}
            if action == "help":
                msg = self.action_help(input)
            elif action == "hello":
                msg['text'] = "Hello {}, nice to talk to you.".format(self.messageconfig.get('firstname',''))
            elif action == 'sensor':
                if 'MARTAS' in self.configuration.get('purpose'):
                    msg['text'] = self.action_martas_sensors(debug=False)
                if 'MARCOS' in self.configuration.get('purpose'):
                    pass
            elif action == 'imbot':
                msg = self.action_imbot(input)
            elif action == 'system':
                msg['text'] = self.action_system()
            elif action == 'martas':
                msg = self.action_martas(input,debug=debug)
            elif action == 'marcos':
                msg = self.action_marcos(input,debug=debug)
            elif action == 'cam':
                msg = self.action_cam(debug=debug)
            elif action == 'status':
                msg['text'] = self.action_status(debug=debug)
            elif action == 'getlog':
                msg['text'] = self.action_getlog(input, debug=debug)
            elif action == 'getdata':
                msg['text'] = self.action_getdata(input, debug=debug)
            elif action == 'plot':
                msg = self.action_plot(input, debug=debug)
            elif action == 'switch':
                msg = self.action_switch(input, debug=debug)
            elif action == 'upload':
                msg = self.action_upload()
            elif action == 'aperta':
                #msg['text'] = self.action_tmate(input)
                msg = self.action_upterm()
            elif action == 'getip':
                msg = self.action_getip(input, debug=debug)
            elif action == 'badwords':
                msg['text'] = self.action_badwords()
            elif action == 'figure1':
                msg = self.action_figure(figure="fig1")
            elif action == 'figure2':
                msg = self.action_figure(figure="fig2")
            elif action == 'reboot':
                msg = self.action_reboot(debug=debug)
            elif action == 'martasupgrade':
                msg = self.action_martasupgrade(debug=debug)
            elif action == 'version':
                msg['text'] = self.commanddict['version'].get('description','')

            adict['message'] = msg
            messagedict[action] = adict
        if debug:
            print (" Obtained the following messages:", messagedict)
        return messagedict

    def action_help(self, printhidden=False):
        """
        DESCRIPTION
            print dictionary of commands
        """
        printall = False
        message = {}
        mesg = ''

        for com in self.commanddict:
            cd = self.commanddict.get(com)
            if 'all' in cd.get('availability') or printall:
                mesg += "COMMAND: *{}*\n".format(com)
                mesg += "{}\n\n".format(cd.get('description'))
                # options = self.extract_options(comdic.get('options'))
            if 'MARTAS' in cd.get('availability') and 'MARTAS' in self.configuration.get('purpose'):
                mesg += "COMMAND: *{}*\n".format(com)
                mesg += "{}\n\n".format(cd.get('description'))
                # options = self.extract_options(comdic.get('options'))
            if 'MARCOS' in cd.get('availability') and 'MARCOS' in self.configuration.get('purpose'):
                mesg += "COMMAND: *{}*\n".format(com)
                mesg += "{}\n\n".format(cd.get('description'))
                # options = self.extract_options(comdic.get('options'))
            if 'IMBOT' in cd.get('availability') and 'IMBOT' in self.configuration.get('purpose'):
                mesg += "COMMAND: *{}*\n".format(com)
                mesg += "{}\n\n".format(cd.get('description'))
                # options = self.extract_options(comdic.get('options'))

        message['text'] = mesg
        return message


    def action_martas_sensors(self, debug=False):
        """
        DESCRIPTION
            provide basic sensorlist and show whether they are active or not
        """
        sensorlist = self.messageconfig.get("sensorlist",[])
        mqttpath = self.messageconfig.get("mqttpath","")
        mesg = "Sensors:\n"
        for s in sensorlist:
            se = s.get('sensorid').replace('$','').replace('?','').replace('!','')
            try:
                lf = _latestfile(os.path.join(mqttpath,se,'*'),date=True)
                diff = (datetime.now(timezone.utc).replace(tzinfo=None)-lf).total_seconds()
                flag = "active"
                if diff > 300:
                    flag = "inactive since {:.1f} hours".format(diff/3600.)
            except:
                flag = "no buffer found"
            mesg += "{}: {}\n".format(se,flag)
        return mesg


    def action_system(self):
        """
        DESCRIPTION
            get system information on hardware and software versions
        """
        mesg = ''
        try:
            import platform
            sysls = platform.uname()
            mesg += "System:\n----------\nName: {}\nOperating system: {} ({})\nKernel: {}\nArchitecture: {}".format(sysls[1],sysls[0],sysls[3],sysls[2],sysls[4])
        except:
            pass
        mesg += "\n\nSoftware versions:\n----------\nMagPy Version: {}".format(magpyversion)
        #mesg += "\nTelegramBot Version: {}".format(tgpar.version)
        mesg += "\nMARTAS/MARCOS Version: {}".format(__version__)
        mesg += "\nPython Version: {}".format(sys.version)

        return mesg


    def action_martas(self, input, debug=False):
        """
        DESCRIPTION:
           issue commend for MARTAS/MARCOS jobs i.e. restart martas process
        """
        message = {}
        if debug:
            print("Running ACTION martas")
            print("-----------------------")
        mainjob = "{}".format(os.path.join(self.configuration.get('base'),'runmartas.sh'))
        if not os.path.isfile(mainjob):
            mesg = "martas: could not find martas call\n"
            return mesg
        command = ""

        comms = ["start","stop","restart","update","status"]
        for jc in comms:
            searchterm = "martas {}".format(jc)
            if searchterm in input:
                command = jc
        if not command:
            if debug:
                print ("Did not find start, stop, restart command - skipping")
            return ""

        call = 'bash {} {} &'.format(mainjob, command)
        message['call'] = [call]
        message['text'] = "Sending MARTAS call: {}\n".format(command)
        return message


    def action_marcos(self, input, debug=False):
        """
        DESCRIPTION:
            restarting marcos process
        """
        message = {}
        command = ""
        cleaninput = ""
        if debug:
            print("Running ACTION marcos")
            print("-----------------------")
        collectorpath = os.path.join(self.configuration.get('base'),"collect-*")
        if debug:
            print (" collector jobs:", collectorpath)
        collectlist = glob.glob(collectorpath)
        if debug:
            print (" collector jobs:", collectlist)

        comms = ["start","stop","restart","status"]
        for jc in comms:
            if jc in input:
                command = jc
        if debug:
            print (" found command", command)
        if not command:
            if debug:
                print (" did not find start, stop, restart command - skipping")
            message['text'] = "did not find start, stop, restart command - skipping"
            return message
        else:
            cleaninput = input.replace("marcos","").replace(command,"").strip()

        if debug:
            print (" found specific brocker", cleaninput)

        # now renmove marcos and startcommand. Check if remaining text corresponds to a collector
        if cleaninput:
            collectlist = [coll for coll in collectlist if coll.find(cleaninput) > -1]

        if debug:
            print (" collector jobs remaining:", collectlist)
        calls = []
        mesg = ""
        if collectlist and len(collectlist)>0:
            for coll in collectlist:
                call = "bash {} {} &".format(coll,command)
                calls.append(call)
                mesg += "sending {} command to {}...\n".format(command,coll)
        message['text'] = mesg
        message['call'] = calls
        return message


    def action_reboot(self, debug=True):
        """
        DESCRIPTION:
            Rebooting the system
            In order to activate:
            $ sudo visudo
            add the following line at the bottom
            $USER ALL=(ALL) NOPASSWD: ALL
            where $USER is your username
        """
        message = {}
        message['call'] = ["sudo /sbin/reboot"]
        message['text'] = "rebooting...\n"
        return message


    def action_martasupgrade(self, debug=True):
        """
        DESCRIPTION:
            updating martas using pip - requires martas 2 to be published on pypi
            ! will not work on system python as sudo and --break-systempackages are required
        """
        message = {}
        message['call'] = ["pip install -U martas", "martas_init -U"] # "pip install -U martas",
        message['text'] = "upgrading martas...\n"
        return message


    def action_cam(self, debug=False):
        """
        DESCRIPTION:
            accessing webcam(s) and taking snapshots
        """
        message = {}
        tmppath = self.messageconfig.get("tmppath","")
        camport = self.configuration.get("camport","")
        camoptions = self.configuration.get("camoptions","")
        usedcamport = None
        calls = []

        if camport:
            usedcamport = self._getcam(camport)

        if not usedcamport:
            mesg = "No camport identified\n"
        else:
            mesg = "Taking snapshot using fswebcam\n"
            if debug:
                print("Creating image...")
                print("Selected cam port: {} and temporary path {}".format(usedcamport, tmppath))
            call = "/usr/bin/fswebcam -d {} {} {}".format(usedcamport, camoptions,
                                                              os.path.join(tmppath, 'webimage.jpg'))
            calls.append(call)

        message['text'] = mesg
        message['call'] = calls
        return message


    def action_figure(self, figure="fig1"):
        message = {}
        message['pictures'] = [self.configuration.get(figure)]
        return message


    def action_status(self, debug=False):
        """
        DESCRIPTION:
            status
        """
        msg = self._getspace()
        if debug:
            print ("Status from getspace", msg)
        msg += self._jobprocess(debug=debug)
        if debug:
            print ("Status from sobprocess", msg)
        return msg


    def action_badwords(self):
        return "Don't be rude.\nI am just a program, acting as mirror now.\n"


    def action_getlog(self, input, debug=False):
        """
        DESCRIPTION:
            get logging information
        """
        logpath = ""
        tmppath = ""
        cmd = input.replace('getlog', '').replace('print log', '').replace('send log', '').replace('get log', '')
        try:
            N = int(re.search(r'\d+', cmd).group())
            cmd = cmd.replace(str(N), '')
        except:
            N = 10
        if not N:
            N = 10

        cmd = cmd.strip()
        syslogfiles = ['syslog', 'dmesg', 'messages', 'faillog']
        telegramlog = self.configuration.get('bot_logging','')
        martaspath = self.configuration.get('martaspath','')
        martaslog = os.path.join(martaspath,"log")

        martaslogfiles = glob.glob(os.path.join(martaslog, '*.log'))
        martaslogfiles = [os.path.basename(ma) for ma in martaslogfiles]
        tmppath = os.path.join(martaslog, "martas.log")

        if debug:
            print ("Logging command looks like", cmd, martaslogfiles)

        if len(cmd) > 3:  # at least three characters remaining
            for logfile in syslogfiles:
                if cmd.find(logfile) > -1:
                    tmppath = os.path.join('/var/log', logfile)
            for logfile in martaslogfiles:
                if logfile.find("{}.".format(cmd)) > -1:
                    tmppath = os.path.join(martaslog, logfile)
            if cmd.find('telegrambot') > -1:
                tmppath = telegramlog
            if debug:
                print(" identified logfile", tmppath)
            if os.path.isfile(tmppath):
                logpath = tmppath
        if not logpath:
            logpath = os.path.join(martaslog, "martas.log")
        if os.path.isfile(logpath):
            if debug:
                print("Checking logfile {}".format(logpath))
            mesg = self._tail(logpath, lines=N)
        else:
            mesg = "getlog:\nlogfile not existing\n"

        return mesg


    def action_switch(self, command, debug=False):
        """
        DESCRIPTION
            send switching command to  a connected microcontroller
            Possible switch commands are:
            swP:0:4   = switch - off - port 4
            swP:1:4   = switch - on - port 4
            swP:1:5   = switch - on - port 4
            swS, Status = get status
        """
        message = {}
        call = ''
        martasconf = self.configuration.get('martasconfig','')
        command = command.replace("switch","").replace("heating ","swP:").replace("on ","1:4").replace("off ","0:4").strip()
        if command.find('swP') > -1:
            pass
        else:
            # Show switch status if no specific switch command is given
            command = "swS"

        try:
            if debug:
                print("Running switch command...")
            python = sys.executable
            #path = '/home/cobs/MARTAS/app/ardcomm.py'
            path = os.path.join(self.configuration.get('martasapp',''),'ardcomm.py')
            if debug:
                print("tpath: {}".format(path))
            option = '-c'
            call = "{} {} {} {} -m {}".format(python,path,option,command,martasconf)
            mesg = "Running command {}\n".format(command)
        except:
            mesg = "martas: check_call problem\n"
        message['call'] = [call]
        message['text'] = mesg
        return message


    # TODO deprecated
    def action_tmate(self, command, debug=False):
        """
        DESCRIPTION
           Open a tmate access for a specific time
           This method is solely to be activated on remote sensor stations in mobile networks.
        REQUIREMENTS
           tmate installation
           tmate api key and a named session (https://tmate.io/#api_key)
           and tmate.conf preconfigured on the remote machine
        """
        # 0. check if psutil and tmate are existing. If not return fail
        try:
            if os.path.isfile("/usr/local/bin/tmate"):
                pass
            else:
                return ("check requirements")
        except:
            return ("check requirements")

        # 1. step check whether tmate is running and stop this process
        def _find_process(name):
            procs = list()
            # Iterate over the all the running process
            for proc in psutil.process_iter():
                try:
                    pid = 0
                    if proc.name() == name:
                        pid = proc.pid
                    if pid:
                        procs.append(pid)
                except:
                    pass
            return procs

        def _kill_processes(pids, debug=False):
            for pid in pids:
                os.kill(pid, signal.SIGTERM)  # or signal.SIGKILL
                if debug:
                    print('killed process with pid: {}'.format(pid))

        try:
            if debug:
                print("Killing existing tmate processes")
            processes = _find_process("tmate")
            _kill_processes(processes, debug=debug)
            if debug:
                print("... done")
        except:
            pass
        # 2. Run tmate accecc - do not broadcast login (pwd needs to be known by user)
        mesg = "Opening secret entrance door....\n"
        try:
            #proc2 = subprocess.Popen(['sudo', '-u', user, 'tmate', '-F', 'new-session'], stdout=subprocess.PIPE)
            proc2 = subprocess.Popen(['tmate', '-F', 'new-session'], stdout=subprocess.PIPE)
            mesg += "success - door open for one session\n"
        except:
            mesg += "failed\n"

        # 3. (optional) start a scheduler to kill tmate process after duration

        return mesg


    def action_imbot(self, input):
        message = {}
        cmd = input.replace('imbot', '').replace('IMBOT', '')
        cmd = cmd.strip()
        call = ""
        year = ""
        imo = ""
        yearl = re.findall(r'\d+', cmd)
        print ("got year", yearl)
        if yearl and len(yearl) > 0:
            for y in yearl:
                cmd = cmd.replace(y,"")
            cmd = cmd.strip()
            if 1900 < int(yearl[-1]) < 2177:
                year = "-y {}".format(int(yearl[-1]))
        if cmd.find('minute') > -1:
            cmd = cmd.replace('minute','')
            cmd = cmd.strip()
            res = 'minute'
        else:
            cmd = cmd.replace('second','')
            cmd = cmd.strip()
            res = 'second'
        print ("remaining command:", cmd)
        if cmd.find('bar') > -1:
            cmd = cmd.replace('bar','')
            cmd = cmd.strip()
            typ = 'bar'
        elif cmd.find('imo') > -1:
            cmd = cmd.replace('imo','')
            cmd = cmd.strip()
            typ = 'imo'
            # get IMO from remaining stream
            print ("remaining imo", cmd, len(cmd))
            if len(cmd) == 3:
                imo = "-i {}".format(cmd.upper())
        else:
            cmd = cmd.replace('list','')
            cmd = cmd.strip()
            typ = 'list'

        imbotconfig = self.configuration.get("imbotconfig","")

        call = "imbot_chart -c {a} -t {b} {c} -r {d} {e}".format(
                a=imbotconfig, b=typ, c=year, d=res, e=imo)
        message['call'] = [call]
        message["text"] = "Requesting IMBOT report of typ {}...\n".format(typ)
        if typ == 'bar':
            message["pictures"] = ['/tmp/bar_levels.png']
        return message


    def action_upload(self):

        message = {}
        calls = []
        configpath = self.configuration.get('uploadconfig','')
        memorypath = self.configuration.get('uploadmemory','')
        if not configpath:
            message['text'] = "Upload not configured - aborting\n"
        else:
            if not memorypath:
                memorypath = "/tmp/martas_tgupload_memory.json"
            python = sys.executable
            path = os.path.join(self.configuration.get('martasapp',''),'file_upload.py')
            optioncfg = '-j'
            optionmem = '-m'
            if configpath:
                call = "{} {} {} {} {} {}".format(python,path,optioncfg,configpath,optionmem,memorypath)
                calls.append(call)
            message['call'] = calls
            message['text'] = "Requesting upload...\n"
        return message


    def action_upterm(self, debug=False):
        """
        DESCRIPTION
           Open a ssh connection to a existing upterm terminal
           This method is solely to be activated on remote sensor stations in mobile networks.
        REQUIREMENTS
           upterm host, which provides secure ssh link
        """
        message = {}
        message['call'] = ["upterm host --accept"]
        message["text"] = "Requesting ssh connection...\n"
        return message
        #import paramiko

        # ssh
        #ssh = paramiko.SSHClient()
        #ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())  # this will automatically add the keys
        #ssh.connect(machineHostName, username=user, password=password)

        pass


    def action_getip(self, input, debug=False):
        """
        DESCRIPTION
           getting the current IP
        """
        message = {}
        calls = []
        call = r"ip -br a"
        calls.append(call)
        message['call'] = calls
        message['text'] = "Requesting IP...\n"
        return message


    def action_plot(self, input, debug=False):
        """
        DESCRIPTION
           plotting subroutine
        """
        mesg = ""
        message = {}
        mqttpath = self.messageconfig.get('mqttpath',"")
        tmppath = self.messageconfig.get('tmppath',"")

        cmd = input
        for word in self.commanddict['plot'].get('commands'):
            cmd = cmd.replace(word, '')
        if debug:
            print (" Plotting...")
            print(" reduced command", cmd)
        sensoridlist = self._identify_sensor(cmd)
        if len(sensoridlist) > 1:
            print("Too many sensors selected - using only {}".format(sensoridlist[0]))
        elif len(sensoridlist) == 0:
            mesg = "You need to specify a sensorid - check 'sensors' to get IDs"
        else:
            sensorid = sensoridlist[0]
            cmd = cmd.replace(sensorid, '')
            # Getting time interval
            cmd = cmd.split()
            # default start and endtime
            endtime = datetime.now(timezone.utc).replace(tzinfo=None)
            starttime = datetime.now(timezone.utc).replace(tzinfo=None) - timedelta(days=1)
            datelist = []
            if len(cmd) >= 1:
                for el in cmd:
                    newdate = self._identify_dates(el)
                    if newdate:
                        datelist.append(newdate)
                if len(datelist) > 0:
                    starttime = min(datelist)
                if len(datelist) > 1:
                    endtime = max(datelist)

            if debug:
                print("Now create plot for", sensorid, starttime, endtime)
            # and save to temporary directory
            data = read(os.path.join(mqttpath, sensorid, '*'), starttime=starttime, endtime=endtime)
            matplotlib.use('Agg')
            f, a = mp.tsplot(data)
            f.savefig(os.path.join(tmppath,'tmp.png'))
            mesg = "Plotting {}\n".format(sensorid)
            message['pictures'] = [os.path.join(tmppath,'tmp.png')]

        message['text'] = mesg
        return message


    def action_getdata(self, command, debug=False):
        """
        DESCRIPTION
           get data
        """
        def CreateSensorMsg(valdict):
            mesg = ''
            for sensor in valdict:
                mesg += "Sensor: {}\n".format(sensor)
                contentdict = valdict.get(sensor)
                keys = contentdict.get('keys')
                for key in keys:
                    keydict = contentdict.get(key)
                    mesg += "  {}: {} {} (key: {})\n".format(keydict.get('element'), keydict.get('value'),
                                                             keydict.get('unit', ''), key)
                    mesg += "  at {}\n".format(contentdict.get('starttime').strftime("%Y-%m-%d %H:%M:%S"))
            return mesg

        cmd = command.replace('data', '').replace('get', '')
        cmdsplit = cmd.split()
        mesg = "Data:\n-----------\n"
        if len(cmdsplit) > 0:
            sensoridlist = self._identify_sensor(cmd)
            # tglogger.info("  found sensors: {}".format(sensoridlist))
            for sensorid in sensoridlist:
                cmd = cmd.replace(sensorid, '')
            if len(sensoridlist) == 0:  # if only dates are provided
                sensoridlist = [None]
            starttime = self._identify_dates(cmd)  # dates is a list
            for sensorid in sensoridlist:
                if debug:
                    print (" Obtaining data for {} starting at {}".format(sensorid, starttime))
                valdict = self.getdata(sensorid=sensorid, starttime=starttime)
                mesg += CreateSensorMsg(valdict)
        else:
            valdict = self.getdata()
            mesg += CreateSensorMsg(valdict)
        return mesg

    def execute_call(self, call, text="", searchcrit="", debug=False):
        """
        DESCRIPTION
            executes the calls as provided within the message dictionaries
            Actions with calls:
            getip ??, upload, imbot, cam, martasupgrade, reboot, switch, martas, marcos
            Methods issueing call already: tmate,
        """
        try:
            if debug:
                print ("Executing call", call)
            if call.find("upterm") > -1:
                # Special treatment required - as terminal needs to be kept open (p.communicate() would close the terminal
                p = subprocess.Popen([call], stdout=subprocess.PIPE, stdin=subprocess.PIPE, shell=True)
                oput = []
                count = 0
                line = p.stdout.readline().decode()
                while count < 10:
                    if line.find("SSH Command") > -1:
                        oput.append(line)
                        break
                    elif line.find("ssh") > -1: #required for new upterm (Jan 2026)
                        oput.append(line+"\n")
                    line = p.stdout.readline().decode()
                    count += 1
                output = "".join(oput)
            elif call.find("martas") > -1 or call.find("collect") > -1:
                # Special treatment required - as terminal needs to be kept open (p.communicate() would close the terminal
                p = subprocess.Popen([call], stdout=subprocess.PIPE, stdin=subprocess.PIPE, shell=True)
                output = "your wish is my command"
            else:
                p = subprocess.Popen(call, stdout=subprocess.PIPE, shell=True)
                (output, err) = p.communicate()
                output = output.decode()
            mesg = output
            print (" ... got:", mesg)
            print (" length:", len(mesg))
            if len(mesg) > 4000:
                mesg = "Message too long - showing only last part:\n\n... " + mesg[-3800:]
        except subprocess.CalledProcessError:
            mesg = " check: call {} did not work".format(call)
        except:
            mesg = "getip: check_call problem"
        return mesg


    def getdata(self, starttime=None, sensorid=None, interval=300, mean='mean'):
        """
        DESCRIPTION
            returns by default a 5 min mean of last values of each sensor
        OPTIONS
            startdate (datetime) : define a specific time
            intervals (int) : define an interval to average values (default one minute)
            mean (string)   : type of average (mean, median)
            sensorid (string) : if no sensorid is given all sensors from sensorlist are used
        RETURN
            a dictionary looking like {'SensorID' : {'keys': 't1,t2,var1',
                                                     't1' : {'element': 'T', 'value': 23.31, 'unit': 'deg C'},
                                                     ...
                                                     }
                                       ...
                                      }
        """
        sensorlist = self.messageconfig.get("sensorlist",[])
        mqttpath = self.messageconfig.get("mqttpath","")

        def GetVals(header, key):
            keystr = "col-{}".format(key)
            element = header.get(keystr, 'unkown')
            unit = header.get('unit-' + keystr, 'arb')
            return element, unit

        if not starttime:
            starttime = datetime.now(timezone.utc).replace(tzinfo=None) - timedelta(seconds=interval)
            endtime = None
        else:
            endtime = starttime + timedelta(seconds=interval)

        senslist = [s.get('sensorid').replace('$', '').replace('?', '').replace('!', '') for s in sensorlist]
        if sensorid and sensorid in senslist:
            senslist = [sensorid]
        returndict = {}
        for s in senslist:
            print(" dealing with sensor:", s)
            print(" between:", starttime, endtime)
            contentdict = {}
            if os.path.isdir(os.path.join(mqttpath, s)):
                try:
                    data = read(os.path.join(mqttpath, s, '*'), starttime=starttime, endtime=endtime)
                except:
                    data = DataStream()
                if len(data) > 0:
                    print(s, data.length(), starttime, endtime)
                    contentdict['keys'] = data._get_key_headers()
                    st, et = data._find_t_limits()
                    contentdict['starttime'] = st
                    contentdict['endtime'] = et
                    print("here", st, et)
                    for key in data._get_key_headers():
                        print(key)
                        valuedict = {}
                        element, unit = GetVals(data.header, key)
                        value = data.mean(key)
                        valuedict['value'] = value
                        valuedict['unit'] = unit
                        valuedict['element'] = element
                        contentdict[key] = valuedict
                    returndict[s] = contentdict
                else:
                    contentdict['keys'] = ['all']
                    contentdict['starttime'] = starttime
                    contentdict['endtime'] = endtime
                    valuedict = {}
                    valuedict['value'] = 'no data within last {} secs'.format(interval)
                    contentdict['all'] = valuedict
                    returndict[s] = contentdict

        return returndict

    def get_pid(self, name):
        """
        DESCRIPTION
            obtain the PID of given call

        An identical method is used marcos_view
        """
        pid = 0
        for proc in psutil.process_iter(attrs=["pid", "name", "exe", "cmdline"]):
            if isinstance(proc.info.get('cmdline'), (list, tuple)):
                for cmd in proc.info.get('cmdline'):
                    if name in cmd:
                        pid = proc.pid
                        break
                if pid:
                    break
        return pid



class TestActionHandler(unittest.TestCase):
    """
    Test environment for all methods
    """

    def test_interpret(self):
        act = ActionHandler()
        test = act.interpret("well: an incredibly loud cry for martas help and logging, but preferebly help")
        self.assertEqual(len(test), 3)

    def test_execute_action(self):
        act = ActionHandler()
        mytext = "hello help sensor imbot system martas marcos cam status getlog getdata plot switch fuck figure1 figure2 reboot martasupgrade getip aperta upload"
        al = act.interpret(mytext)
        msg = act.execute_action(al, input=mytext, debug=True)


    def test_action_martas(self):
        act = ActionHandler()
        mytext = ("martas restart")
        al = act.interpret(mytext)
        msg = act.execute_action(al, input=mytext, debug=True)
        td = msg.get('martas',{})
        self.assertTrue(isinstance(td['message'].get('call'), (list,tuple)))

    def test_action_marcos(self):
        act = ActionHandler()
        mytext = ("marcos janus restart")
        al = act.interpret(mytext)
        msg = act.execute_action(al, input=mytext, debug=True)
        td = msg.get('marcos',{})
        self.assertTrue(isinstance(td['message'].get('call'), (list,tuple)))

    def test_action_getlog(self):
        act = ActionHandler()
        mytext = ("getlog martas 20")
        al = act.interpret(mytext)
        msg = act.execute_action(al, input=mytext, debug=True)
        td = msg.get('getlog',{})
        self.assertTrue(td['message'].get('text'))

    def test_action_getdata(self):
        act = ActionHandler()
        mytext = ("data TEST_1234_0001")
        al = act.interpret(mytext)
        msg = act.execute_action(al, input=mytext, debug=True)
        td = msg.get('getdata',{})
        print ("GET data", td)
        self.assertTrue(td['message'].get('text'))

    def test_zaction_plot(self):
        act = ActionHandler()
        mytext = ("plot TEST_1234_0001 2025-07-31")
        al = act.interpret(mytext)
        msg = act.execute_action(al, input=mytext, debug=True)
        td = msg.get('plot',{})
        print ("Plotting", td)
        #self.assertTrue(td['message'].get('text'))


    def test_execute_call(self):
        act = ActionHandler()
        #upload, imbot, cam, martasupgrade, reboot, switch, martas, marcos
        callcommands = {"getip" : "getip wlp0s20f3"} #, "marcos" : "marcos local restart"}
        for comm in callcommands:
            fc = callcommands.get(comm)
            lines = []
            msg = ""
            al = act.interpret(fc)
            msg = act.execute_action(al, input=fc, debug=True)
            td = msg.get(comm,{})
            print ("Call command:", td['message'].get('call'))
            calls = td['message'].get('call',[])
            for call in calls:
                msg = act.execute_call(call)
                if msg:
                    lines.append(msg)
            if lines:
                msg = "\n".join(lines)
            if msg:
                print ("Result for {}: Obtained {}".format(comm,msg))


if __name__ == "__main__":
    unittest.main(verbosity=2)


"""
ad7714protocol version 1.1
for board v0.1 and v0.2

Purpose:
    reads data from an AD7714 (Analog Devices)

Requirements:
    SPI interface on a RaspberryPi3
    
Called by:
    acquisition from magpy
"""

from __future__ import print_function

############    user definitions    ############

# select GAIN between 0 and 7
#  nr    amplification
#   0 .. 1
#   1 .. 2
#   2 .. 4
#   ......
#   6 .. 64
#   7 .. 128

GAIN = 1

###### please don't edit beyond this line ######


import struct
from datetime import datetime, timezone

from twisted.python import log
from martas.core import methods as mm
import time

# Raspberry Pi specific
try:
    import spidev
    import RPi.GPIO as GPIO
except:
    log.msg('sorry, equipment not prepared to communicate over SPI')
    raise

# TODO brauch ich das? und wenn, ab ins acs!
def datetime2array(t):
    return [t.year,t.month,t.day,t.hour,t.minute,t.second,t.microsecond]

print("setup comm interface to AD7714:")
print("GPIO warnings are not disabled intentionally")
# set GPIO mode to pinnumbers
GPIO.setmode(GPIO.BOARD)

# board specific settings
# define pin32 as input for /DRDY
DRDY=32
GPIO.setup(DRDY,GPIO.IN)
# define pin7 as output for /RESET
RESET=7
GPIO.setup(RESET,GPIO.OUT,initial=GPIO.HIGH)
# next line not needed anymore - 'initial' prevents a reset when starting a second instance of ad7714protocol
#GPIO.output(RESET,GPIO.HIGH)

# setup spi
SPI=spidev.SpiDev()
SPI.open(0, 0)
# no difference of SPI speed for RaspberryPi and AD7714
# makes sense when using an oszilloscope
# RPi3B+ supports higher speeds, limiting for proper transmission
SPI.max_speed_hz=500000
print("max_speed_hz", SPI.max_speed_hz)

# AD7714

# specific SPI settings
# CPOL 0, CPHA 1 
SPI.mode=1

# board specific settings
# AD7714's system clock
# *** board v0.1 ***
#CLK=1000000
# *** board v0.2 ***
CLK=2457600
print("system clock", CLK)
# *** board v0.1 ***
# using channel 5: differential input AIN3-AIN4 (pin9-pin10)
#CHANNEL = 5
# *** board v0.2 ***
# using channel 4: differential input AIN1-AIN2 (pin7-pin8)
CHANNEL = 4
# wordlength 16bit:0 24bit:1
BIT = 1
# Filter sets sampling rate. Depends on system clock
FILTER = 0x0c0

# init watchdog
watchdog = {}
watchdog['count_repetitions'] = 0
watchdog['init_max_rep'] = 20
watchdog['max_repetitions'] = watchdog['init_max_rep']
watchdog['oldvalue'] = 999999

global int_comm
int_comm = ""

def reset():
    """
    reset AD7714
    AD7714 requires minimum 100ns on the /RESET pin
    RPi's sleep is not so fast anyhow...
    """
    GPIO.output(RESET,GPIO.LOW)
    # time.sleep(0.0000001)
    time.sleep(0.01)
    GPIO.output(RESET,GPIO.HIGH)
    
def zo(r):
    """
    used by txreg and rxreg
    return value should there be [0] once
    if not, some sync is wrong
    """
    if not r==[0]:
        print("NOT ZERO!, but",r)

def zt(r):
    """
    used by txreg
    return value should there be 0 three times
    if not, some sync is wrong
    """
    if not r==[0,0,0]:
        print("NOT ZERO!, but",r)

def txreg(register,channel,value):
    # write to a register on AD7714
    x=(register<<4)|channel
    # send which register is next
    zo(SPI.xfer([x]))
    # write to register
    if not register in [6,7]:
        # transmit 1 byte
        zo(SPI.xfer([value]))
    else:
        # transmit 3 byte (24bit)
        v=[(value>>16)&0xff,(value>>8)&0xff,value&0xff]
        zt(SPI.xfer(v))

def rxreg(register,channel):
    # read from a register
    if register in [1,2,3,4]:
        nr_bytes=1
    if register==5:
        # WL 0:16bit 1:24bit
        nr_bytes=2+WL
    if register in [6,7]:
        nr_bytes=3
    x=(register<<4)|8|channel
    zo(SPI.xfer([x]))
    r=SPI.readbytes(nr_bytes)
    return r

def setMode(channel=4,MD=0):
    """
    set mode: refer to AD7714s manual
    """
    # read unchanged values from mode register
    vm=rxreg(1,channel)
    vm=(vm[0]&0x1f)|(MD<<5)
    txreg(1,channel,vm)

def setGain(channel=4,G=0):
    """
    set gain of AD7714s preamp in powers of two
    amplification = 2 ** G
    """
    # read unchanged values from mode register
    vm=rxreg(1,channel)
    vm=(vm[0]&0xe3)|(G<<2)
    txreg(1,channel,vm)

def setWL(wordlength=None,channel=4):
    """
    set wordlength
    WL 0:16bit, 1:24bit
    ensure not to disturb SPI communication!
    (not yet implemented)
    """
    #GPIO.wait_for_edge(DRDY, GPIO.RISING)
    # TODO war naechste Zeile Grund fuer Probleme?
    #GPIO.wait_for_edge(DRDY, GPIO.FALLING)
    # read unchanged values from filter high register
    vf=rxreg(2,channel)
    global WL
    WL=(vf[0]>>6)&0x1
    if wordlength is None:
        # WL is set, no need to change registers
        pass
    else:
        if wordlength in [0,1]:
            # write to WL bit on AD7714
            vf=(vf[0]&0xbf)|(wordlength<<6)
            txreg(2,channel,vf)
            print('...changing WL to ',wordlength)
            vf=rxreg(2,channel)
            WL=(vf[0]>>6)&0x1
            if not WL==wordlength:
                print("couldn't change wordlength!")
        else:
            print('wordlength must be 0 or 1')
    print('Wordlength is',(2+WL)*8,'bit')

def setFilter(FS=0xfa0,channel=4):
    """
    set filter: refer to AD7714s manual
    """
    # read unchanged values from filter high register
    vf=rxreg(2,channel)
    vf=(vf[0]&0xf0)|(FS>>8)
    txreg(2,channel,vf)
    # write to filter low register
    vf=FS&0xff
    txreg(3,channel,vf)

def calcSamp2Filt(samplingrate):
    """
    a small calculator to help the user in python's interpreter mode
    careful in normal operation, not every sampling rates are supported
    """
    filt=CLK/128./samplingrate
    print("FS has to be",filt,"rounded",round(filt))
    print("in hex:",hex(int(round(filt))))
    if round(filt)<19:
        print("...but this is too low!")
    elif round(filt)>4000:
        print("...but this is too high!")
    else:
        print("the resulting sampling rate is ",CLK/128./int(round(filt)))
    return int(round(filt))    

def info():
    """
    lists content of AD7714's registers
    for test purposes
    """
    print("registers of AD7714:")
    print()
    mode=rxreg(1,0)
    MD=(mode[0]>>5)&7
    G=(mode[0]>>2)&7
    BO=(mode[0]>>1)&1
    FSYNC=(mode[0])&1
    print("mode register\t\tvalue\t",hex(mode[0]))
    print("mode\t\t\tMD\t",MD)
    print("gain\t\t\tG\t",G)
    print("\tmultiplier",2**G)
    print("burnout current\t\tBO\t",BO)
    print("filter synchronization\tFSYNC\t",FSYNC)
    print()
    filth=rxreg(2,0)
    filtl=rxreg(3,0)
    BU=(filth[0]>>7)&1
    WL=(filth[0]>>6)&1
    BST=(filth[0]>>5)&1
    CLKDIS=(filth[0]>>4)&1
    FS=((filth[0]&0xf)<<8)|filtl[0]
    print("filter high register\tvalue\t",hex(filth[0]))
    print("filter low register\tvalue\t",hex(filtl[0]))
    print("bi/unipolar operation\t/B_U\t",BU)
    print("word length\t\tWL\t",WL)
    print("\tword length",16+8*WL)
    print("master clk disable\tCLKDIS\t",CLKDIS)
    print("filter selection\tFS\t",FS)
    print("\tfirst notch freq",(CLK/128./FS),"Hz")
    print()
    print("zero-scale calibration registers")
    zs=[]
    for i in range(3):
        zs.append(rxreg(6,i))
        zs[i]=(zs[i][0]<<16)+(zs[i][1]<<8)+zs[i][2]
        print("\t",i,"\t",zs[i],"\t",hex(zs[i]))
    print()
    print("full-scale calibration registers")
    fs=[]
    for i in range(3):
        fs.append(rxreg(7,i))
        fs[i]=(fs[i][0]<<16)+(fs[i][1]<<8)+fs[i][2]
        print("\t",i,"\t",fs[i],"\t",hex(fs[i]))

def mySettings():
    """
    used in __init__ of class AD7714Protocol
    and for test purposes
    """
    # using channel 5: differential input AIN3-AIN4 (pin9-pin10)
    # setGain: G = 2^(given value) e.g. setGain(channel=5,G=1) results in an amplification of 2^1 = 2
    setGain(CHANNEL,GAIN)
    # setWL: wordlength 0:16bit, 1:24bit
    setWL(BIT)
    # setFilter(5,0xfa0) defines the sampling rate, in this example the minimal
    # setFilter depends on system clock, see calcSamp2Filt
    setFilter(FILTER)

def myCalibration():
    """
    zero calibration removes offset
    """
    # "self calibration" 
    # mode 1, see AD7714 manual, zero scale and full scale cal internally 
    # - apparently not good enough
    #setMode(5,1)
    # mode 5: permanent calibration 
    # - not so "beautiful", similar apearance not before setting double sampling rate 
    #setMode(5,5)
    # mode 4: zero scale cal from input (should be stable), full scale cal internally
    # - should be best, but mode 2 is better
    # mode 2: only zero scale cal from input (should be stable)
    # - best for offset
    setMode(CHANNEL,2)
    # doing zero scale cal again, especially when the sensor is not locked
    time.sleep(0.3)
    setMode(CHANNEL,2)
    # immediate communication would crash settings --> TODO better sync!
    time.sleep(0.01)

def interruptRead(s):
    """
    interrupt routine of class AD7714Protocol
    triggered by AD7714 /DRDY signal
    """
    # at first get the time...
    currenttime = datetime.now(timezone.utc).replace(tzinfo=None)
    # read from data register
    arrvalue=rxreg(5,CHANNEL)
    if len(arrvalue)==2:
        # 16 -> 24bit
        arrvalue.append(0)
    intvalue=(arrvalue[0]<<16) | (arrvalue[1]<<8) | arrvalue[2]
    voltvalue=float(intvalue)/2**24*5-2.5
    # mV better for display
    voltvalue=voltvalue*1000
    
    # TIME TO COMMUNICATE!
    global int_comm
    if int_comm == "mySettings":
        mySettings()
        int_comm = "ok"
    if int_comm == "myCalibration":
        myCalibration()
        int_comm = "ok"
    if int_comm == "info":
        info()
        int_comm = "ok"
    

    # watchdog
    global watchdog
    global Objekt
    if watchdog['oldvalue'] == 999999:
        print('watchdog active')
    if watchdog['oldvalue'] == intvalue:
        watchdog['count_repetitions'] = watchdog['count_repetitions'] + 1
    else:
        if watchdog['count_repetitions'] > 5:
            # avoid a lot of log entries - filter double and triple values in a row
            print('watchdog resetted, count_repetitions:',watchdog['count_repetitions'],'oldvalue:',watchdog['oldvalue'],'intvalue:',intvalue)
        watchdog['count_repetitions'] = 0
        watchdog['max_repetitions'] = watchdog['init_max_rep']
    if watchdog['count_repetitions'] == watchdog['max_repetitions']:
        # probably hung up, too many same values
        print('watchdog ad7714protocol:')
        print('  ',watchdog['max_repetitions'],'same values (intvalue:',intvalue,') in one row - hung up?')
        print('  trying to reset AD7714...')
        # sending LOW to /RESET pin
        reset()
        time.sleep(0.01)
        # loading settings
        mySettings()
        # zero calibration
        myCalibration()
        watchdog['max_repetitions'] = watchdog['max_repetitions'] * 2
        watchdog['count_repetitions'] = 0
    watchdog['oldvalue'] = intvalue
    packcode = "6hLl"
    sensorid = Objekt.sensordict['sensorid']
    header = "# MagPyBin %s %s %s %s %s %s %d" % (sensorid,'[var1]','[U]','[mV]','[1000]',packcode,struct.calcsize(packcode))
    #timestamp=datetime.strftime(currenttime, "%Y-%m-%d %H:%M:%S.%f")
    timestamp = datetime2array(currenttime)
    darray = timestamp
    darray.append(int(round(voltvalue*1000)))


    # TO FILE 
    data_bin = struct.pack(packcode,*darray)
    filedate = datetime.strftime(datetime(darray[0],darray[1],darray[2]), "%Y-%m-%d")
    if not Objekt.confdict.get('bufferdirectory','') == '':
        mm.data_to_file(Objekt.confdict.get('bufferdirectory'), sensorid, filedate, data_bin, header)

    # VIA MQTT
    # instead of external program file TODO: better!
    #def sendData(self, sensorid, data, head, stack=None):
    #sendData.sendData(Objekt,sensorid, ','.join(list(map(str,darray))), header)
    data=','.join(list(map(str,darray)))
    head=header
    # TODO: implement stack correctly!
    stack=1

    topic = Objekt.confdict.get('station') + '/' + sensorid
    senddata = False
    if not stack:
        stack = int(Objekt.sensordict.get('stack'))
    coll = stack

    if coll > 1:
        Objekt.metacnt = 1 # send meta data with every block
        if Objekt.datacnt < coll:
            Objekt.datalst.append(data)
            Objekt.datacnt += 1
        else:
            senddata = True
            data = ';'.join(Objekt.datalst)
            Objekt.datalst = []
            Objekt.datacnt = 0
    else:
        senddata = True

    if senddata:
            if Objekt.count == 0:
                # get all values initially from the database
                #add = "SensoriD:{},StationID:{},DataPier:{},SensorModule:{},SensorGroup:{},SensorDecription:{},DataTimeProtocol:{}".format( sensorid, self.confdict.get('station',''),self.sensordict.get('pierid',''), self.sensordict.get('protocol',''),self.sensordict.get('sensorgroup',''),self.sensordict.get('sensordesc',''), self.sensordict.get('ptime','') )
                #self.client.publish(topic+"/dict", add, qos=self.qos)
                Objekt.client.publish(topic+"/meta", head, qos=Objekt.qos)
                if Objekt.debug:
                    log.msg("  -> DEBUG - Publishing meta --", topic, head)
            Objekt.client.publish(topic+"/data", data, qos=Objekt.qos)
            if Objekt.debug:
                log.msg("  -> DEBUG - Publishing data")
            Objekt.count += 1
            if Objekt.count >= Objekt.metacnt:
                Objekt.count = 0



def readData(channel=4):
    """
    only for test purposes (i.e. in def show)
    not used in normal mode
    """
    GPIO.wait_for_edge(DRDY, GPIO.FALLING)
    return rxreg(5,channel)

def show(channel=4,samples=1):
    """
    shows some samples, for test purposes
    """
    values=[]
    for i in range(samples):
        arrvalue=readData(channel)
        if len(arrvalue)==2:
            # 16 -> 24bit
            arrvalue.append(0)
        print(arrvalue)
        intvalue=(arrvalue[0]<<16) | (arrvalue[1]<<8) | arrvalue[2]
        print(intvalue)
        voltvalue=float(intvalue)/2**24*5-2.5
        print(voltvalue)
        values.append(voltvalue)
    return(values)

def stat(values):
    """
    for test purposes, plausibility of values
    
    """
    mean=0
    for v in values:
        mean=mean+v
    mean=mean/len(values)
    var=0
    for v in values:
        var=var+(v-mean)**2
    var=var/len(values)
    print("mean value",mean)
    print("variance",var)

def seeDRDY():
    """
    only for test purposes
    """
    t=time.time()+1
    i=0
    while time.time()<t:
        GPIO.wait_for_edge(DRDY, GPIO.FALLING)
        i = i + 1
    print("difftime: ",t-time.time())
    print("samples per second: ",i)


# word length should be set in __init__ of class AD7714Protocol
# WL=0: 16bit, WL=1: 24bit
# here WL is set by setWL reading the AD7714 register (default after reset is 0)
# this guaranties that the global variable WL has the correct value
setWL()
# alternatively calculate FILTER from the sampling rate, careful!
# this will override the setting from before:
#FILTER = calcSamp2Filt(15)
print("setup done")



class ad7714Protocol():
    """
    The AD7714 protocol for reading from the AD converter over spi
    """

    def __init__(self, client, sensordict, confdict):

        # set defaults for mqtt
        #sendData.setDefaults(self, client, sensordict, confdict)
        self.qos = 0
        self.debug = False
        self.client = client
        self.sensordict = sensordict
        self.confdict = confdict
        # variables for broadcasting via mqtt:
        self.count=0
        self.datalst = []
        self.datacnt = 0
        self.metacnt = 10


        # reset AD7714
        print('AD7714Protocol: resetting AD7714...')
        reset()
        time.sleep(0.1)


        # display first samples
        # TODO too dangerous:
        #show(channel=CHANNEL,samples=3)
        # due to difficulties setting up an interrupt routine within this object...
        # TODO avoid global variables
        global Objekt
        Objekt=self
#       GPIO.add_event_detect(DRDY, GPIO.FALLING, callback = AD7714Protocol.interruptReadObj(self))
        GPIO.add_event_detect(DRDY, GPIO.FALLING, callback = interruptRead)
        
        # *** indirectly TODO make it better!
        # load settings, zero calibration
        #mySettings()
        global int_comm
        int_comm = "mySettings"
        while not int_comm == "ok":
            time.sleep(0.001)
        # zero calibration
        #myCalibration()
        int_comm = "myCalibration"
        while not int_comm == "ok":
            time.sleep(0.001)
        # display AD7714s register values
        #info()
        #int_comm = "info"
        while not int_comm == "ok":
            time.sleep(0.001)
        # ***

        print("connection to AD7714 via SPI initialized")

